#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import calendar
import copy
import csv
import datetime
import json
import random
import sys
import uuid


COMPUTE = {
    "type": "compute",
    "desc": {},
    "vol": {
        "qty": 1,
        "unit": "instance"}}

COMPUTE_RESOURCE = {
    "availability_zone": "nova",
    "flavor": "m1.nano",
    "image_id": "f5600101-8fa2-4864-899e-ebcb7ed6b568",
    "memory": "64",
    "metadata": {
        "farm": "prod"},
    "name": "prod1",
    "project_id": "f266f30b11f246b589fd266f85eeec39",
    "user_id": "55b3379b949243009ee96972fbf51ed1",
    "vcpus": "1"}

IMAGE = {
    "type": "image",
    "desc": {},
    "vol": {
        "qty": 214106112.0,
        "unit": "B"}}

IMAGE_RESOURCE = {
    "name": "cirros-0.3.4-x86_64-uec-ramdisk",
    "checksum": "be575a2b939972276ef675752936977f",
    "disk_format": "ari",
    "protected": "False",
    "container_format": "ari",
    "min_disk": "0",
    "is_public": "True",
    "min_ram": "0",
    "project_id": "f1873b13951542268bf7eed7cf971e52",
    "resource_id": "08017fbc-b13a-4d8d-b002-4eb4eff54cd4",
    "source": "openstack",
    "user_id": "None",
    "size": "214106112"}

VOLUME = {
    "type": "volume",
    "desc": {},
    "vol": {
        "qty": 1,
        "unit": "GB"}}

VOLUME_RESOURCE = {
    'instance_uuid': 'None',
    'status': 'available',
    'display_name': 'test-vol',
    'event_type': 'volume.create.end',
    'availability_zone': 'nova',
    'tenant_id': 'cd27b013b9db4f4099e273e4b9949023',
    'created_at': '2015-04-28 13:34:25',
    'snapshot_id': 'None',
    'volume_type': '314150bc-221f-4676-a7cf-16f12850b217',
    'host': 'volume.devstack@lvmdriver-1#lvmdriver-1',
    'replication_driver_data': 'None',
    'replication_status': 'disabled',
    'volume_id': '2bed6a3d-468a-459b-802b-44930016c0a3',
    'replication_extended_status': 'None',
    'user_id': '2524d5a52ce64a569d131d7dc1dfb455',
    'launched_at': '2015-04-28 13:34:26.869928',
    'size': '1',
    "project_id": "f1873b13951542268bf7eed7cf971e52",
    "resource_id": "08017fbc-b13a-4d8d-b002-4eb4eff54cd4",
    "source": "openstack",
    "user_id": "None"}

NETWORK_BW_IN = {
    "type": "network.bw.in",
    "desc": {},
    "vol": {
        "qty": 4546.0,
        "unit": "B"}}

NETWORK_BW_OUT = {
    "type": "network.bw.out",
    "desc": {},
    "vol": {
        "qty": 50.0,
        "unit": "MB"}}

NETWORK_BW_RESOURCE = {
    'instance_id': 'eef9673d-5d24-43fd-89f5-2929acc7e193',
    'instance_type': '42',
    'mac': 'fa:16:3e:dd:b2:80',
    'fref': 'None',
    'name': 'tap12a7d4e1-fc',
    "project_id": "f1873b13951542268bf7eed7cf971e52",
    "resource_id": "08017fbc-b13a-4d8d-b002-4eb4eff54cd4",
    "source": "openstack",
    "user_id": "None"}

FLOATING = {
    "type": "network.floating",
    "desc": {},
    "vol": {
        "qty": 1.0,
        "unit": "ip"}}

FLOATING_RESOURCE = {
    'router_id': '3d9b2725-fc90-43d0-b119-630e80e1ec51',
    'status': 'DOWN',
    'event_type': 'floatingip.update.end',
    'tenant_id': 'cd27b013b9db4f4099e273e4b9949023',
    'floating_network_id': '14198fb4-dc96-45fb-9dde-546f6b0f892f',
    'host': 'network.devstack',
    'fixed_ip_address': '10.0.0.5',
    'floating_ip_address': '172.24.4.3',
    'port_id': '12a7d4e1-fcc3-4a3c-8e57-c4baf7787b57',
    "project_id": "cd27b013b9db4f4099e273e4b9949023",
    "resource_id": "ebf6485d-7f6f-4c67-97f7-7896324e12d4",
    "source": "openstack",
    "user_id": "7319b5d1269d4166a402868b570aad19",
    'id': 'ebf6485d-7f6f-4c67-97f7-7896324e12d4'}


class VariationMapper(object):
    day_map = {
        0: 'mon',
        1: 'tue',
        2: 'wed',
        3: 'thu',
        4: 'fri',
        5: 'sat',
        6: 'sun'}
    var_map = {
        'mon': {},
        'tue': {},
        'wed': {},
        'thu': {},
        'fri': {},
        'sat': {},
        'sun': {}}

    def __init__(self, default=1.0):
        self.default = default

    def get_var(self, dt):
        weekday = self.day_map[dt.weekday()]
        if weekday in self.var_map:
            wday_map = self.var_map[weekday]
            if dt.hour in wday_map:
                return wday_map[dt.hour]
            elif 'default' in wday_map:
                return wday_map['default']
        elif 'default' in self.var_map:
            return self.var_map['default']
        return self.default


class VolumeVariationMapper(VariationMapper):
    def get_vol(self, dt):
        value = self.get_var(dt)
        var_value = value * 0.1
        return random.gauss(value, var_value)


class BaseGenerator(object):
    base_sample = None
    base_resource = None
    rand = True
    field_maps = {}

    def __init__(self, nb_res=1, var_map=None, vol_map=None):
        self.nb_res = nb_res
        self.var_map = var_map if var_map else VariationMapper()
        self.vol_map = vol_map
        self.init_mapper()
        self.resources = []

    def init_mapper(self):
        pass

    def generate_resources(self):
        for i in range(self.nb_res):
            res = copy.deepcopy(self.base_resource)
            for field, mapping in self.field_maps.items():
                if hasattr(self, mapping):
                    mapping = getattr(self, mapping)
                if isinstance(mapping, dict):
                    if self.rand:
                        value = random.choice(mapping.keys())
                    else:
                        value = mapping.keys()[i]
                    res[field] = value
                    for k, v in mapping[value].items():
                        res[k] = v
                elif isinstance(mapping, list):
                    if self.rand:
                        value = random.choice(mapping)
                    else:
                        value = mapping[i]
                    res[field] = value
                elif callable(mapping):
                    res[field] = mapping(i)
                else:
                    res[field] = mapping
            self.resources.append(res)

    def generate_samples(self, dt):
        samples = []
        res_var = int(self.var_map.get_var(dt))
        for i in range(res_var):
            sample = copy.deepcopy(self.base_sample)
            sample['desc'] = self.resources[i]
            if self.vol_map:
                qty = self.vol_map.get_vol(dt)
                sample['vol']['qty'] = qty
            elif 'size' in sample['desc']:
                sample['vol']['qty'] = sample['desc']['size']
            # Packing
            sample['desc'] = json.dumps(self.resources[i])
            sample['vol'] = json.dumps(sample['vol'])

            samples.append(sample)
        return samples


class ComputeVarMapper(VariationMapper):
    var_map = {
        'mon': {
            'default': 1.0,
            12: 2.0,
            13: 3.0,
            14: 2.0,
            18: 2.0,
            19: 3.0,
            20: 4.0,
            21: 4.0,
            22: 3.0,
            23: 2.0,
            },
        'tue': {
            'default': 1.0,
            12: 2.0,
            13: 3.0,
            14: 2.0,
            18: 2.0,
            19: 3.0,
            20: 4.0,
            21: 4.0,
            22: 3.0,
            23: 2.0,
            },
        'wed': {
            'default': 1.0,
            12: 2.0,
            13: 3.0,
            14: 2.0,
            18: 2.0,
            19: 3.0,
            20: 4.0,
            21: 4.0,
            22: 3.0,
            23: 2.0,
        },
        'thu': {
            'default': 1.0,
            12: 2.0,
            13: 3.0,
            14: 2.0,
            18: 2.0,
            19: 3.0,
            20: 4.0,
            21: 4.0,
            22: 3.0,
            23: 2.0,
        },
        'fri': {
            'default': 1.0,
            12: 2.0,
            13: 3.0,
            14: 2.0,
            18: 2.0,
            19: 3.0,
            20: 4.0,
            21: 4.0,
            22: 3.0,
            23: 2.0,
        },
        'sat': {
            'default': 2.0,
            12: 3.0,
            13: 4.0,
            14: 3.0,
            18: 3.0,
            19: 4.0,
            20: 4.0,
            21: 4.0,
            22: 4.0,
            23: 3.0,
        },
        'sun': {
            'default': 2.0,
            12: 3.0,
            13: 4.0,
            14: 3.0,
            18: 3.0,
            19: 4.0,
            20: 4.0,
            21: 4.0,
            22: 4.0,
            23: 3.0,
        }}


class ComputeGenerator(BaseGenerator):
    base_sample = COMPUTE
    base_resource = COMPUTE_RESOURCE
    field_maps = {'flavor': 'flavors',
                  'image': 'images',
                  'name': 'generate_name',
                  'resource_id': 'res_id'}

    def init_mapper(self):
        self.flavors = {
            'm1.nano': {
                'vcpus': '1',
                'memory': '64'},
            'm1.micro': {
                'vcpus': '1',
                'memory': '128'}}
        self.images = []
        self.res_id = []
        for i in range(self.nb_res):
            self.res_id.append(str(uuid.uuid1()))

    def generate_name(self, *args):
        basename = 'instance{}'
        return basename.format(args[0])


class ImageGenerator(BaseGenerator):
    base_sample = IMAGE
    base_resource = IMAGE_RESOURCE
    field_maps = {'name': 'images'}
    rand = False

    def init_mapper(self):
        self.images = {
            'cirros-0.3.4-x86_64-uec-kernel': {
                'checksum': '836c69cbcd1dc4f225daedbab6edc7c7',
                'disk_format': 'ari',
                'container_format': 'ari',
                'size': '4969360',
                'resource_id': '5dd34048-6eeb-4b6c-aa51-62487733e5a1'},
            'cirros-0.3.4-x86_64-uec-ramdisk': {
                'checksum': '68085af2609d03e51c7662395b5b6e4b',
                'disk_format': 'aki',
                'container_format': 'aki',
                'size': '3723817',
                'resource_id': 'e512a97d-1ed5-4b27-a55a-1b9e5087936a'},
            'Fedora-x86_64-20-20131211.1-sda': {
                'checksum': '51bc16b900bf0f814bb6c0c3dd8f0790',
                'disk_format': 'qcow2',
                'container_format': 'bare',
                'size': '214106112',
                'resource_id': '3ee99f3f-7ecf-47b2-9a40-6df2d66ef5ae'}}


class VolumeGenerator(BaseGenerator):
    base_sample = VOLUME
    base_resource = VOLUME_RESOURCE
    field_maps = {'volume_id': 'volumes'}
    rand = False

    def init_mapper(self):
        self.volumes = {
            '2bed6a3d-468a-459b-802b-44930016c0a3': {
            'size': '10'},
            '4fd33321-6a5f-4351-94ca-db398cd708e9': {
            'size': '20'}}

    def generate_name(self, *args):
        basename = 'volume{}'
        return basename.format(args[0])


class NetBWVolMapper(VolumeVariationMapper):
    var_map = {
        'mon': {
            'default': 1024,
            12: 2048,
            13: 3072,
            14: 2048,
            18: 2048,
            19: 3072,
            20: 4096,
            21: 4096,
            22: 3072,
            23: 2048,
            },
        'tue': {
            'default': 1024,
            12: 2048,
            13: 3072,
            14: 2048,
            18: 2048,
            19: 3072,
            20: 4096,
            21: 4096,
            22: 3072,
            23: 2048,
            },
        'wed': {
            'default': 1024,
            12: 2048,
            13: 3072,
            14: 2048,
            18: 2048,
            19: 3072,
            20: 4096,
            21: 4096,
            22: 3072,
            23: 2048,
        },
        'thu': {
            'default': 1024,
            12: 2048,
            13: 3072,
            14: 2048,
            18: 2048,
            19: 3072,
            20: 4096,
            21: 4096,
            22: 3072,
            23: 2048,
        },
        'fri': {
            'default': 1024,
            12: 2048,
            13: 3072,
            14: 2048,
            18: 2048,
            19: 3072,
            20: 4096,
            21: 4096,
            22: 3072,
            23: 2048,
        },
        'sat': {
            'default': 2048,
            12: 3072,
            13: 4096,
            14: 3072,
            18: 3072,
            19: 4096,
            20: 4096,
            21: 4096,
            22: 4096,
            23: 3072,
        },
        'sun': {
            'default': 2048,
            12: 3072,
            13: 4096,
            14: 3072,
            18: 3072,
            19: 4096,
            20: 4096,
            21: 4096,
            22: 4096,
            23: 3072,
        }}


class NetworkBWGenerator(BaseGenerator):
    base_sample = NETWORK_BW_IN
    base_resource = NETWORK_BW_RESOURCE
    field_maps = {'instance_id': 'instances',
                  'name': 'generate_name',
                  'mac': 'generate_mac',
                  'resource_id': 'res_id'}
    rand = False

    def init_mapper(self):
        self.instances = []
        self.res_id = []
        for i in range(self.nb_res):
            self.res_id.append(str(uuid.uuid1()))

    def generate_name(self, *args):
        basename = 'tap{}-fc'
        return basename.format(args[0])

    def generate_mac(self, *args):
        basemac = 'fa:16:3e:{:0=2x}:{:0=2x}:{:0=2x}'
        return basemac.format(
            random.randint(1, 255),
            random.randint(1, 255),
            random.randint(1, 255))

    def generate_samples(self, dt):
        samples = []
        self.base_sample = NETWORK_BW_OUT
        samples.extend(super(NetworkBWGenerator, self).generate_samples(dt))
        self.base_sample = NETWORK_BW_IN
        samples.extend(super(NetworkBWGenerator, self).generate_samples(dt))
        return samples


class FloatingGenerator(BaseGenerator):
    base_sample = FLOATING
    base_resource = FLOATING_RESOURCE
    field_maps = {'fixed_ip_address': 'generate_ip_addr',
                  'floating_ip_address': 'generate_floating_addr',
                  'port_id': 'generate_port_id',
                  'resource_id': 'res_id',
                  'id': 'res_id'}
    rand = False

    def init_mapper(self):
        self.res_id = []
        for i in range(self.nb_res):
            self.res_id.append(str(uuid.uuid1()))

    def generate_name(self, *args):
        basename = 'volume{}'
        return basename.format(args[0])

    def generate_port_id(self, *args):
        return str(uuid.uuid1())

    def generate_ip_addr(self, *args):
        baseip = '10.0.0.{}'
        return baseip.format(random.randint(5, 250))

    def generate_floating_addr(self, *args):
        baseip = '172.24.4.{}'
        return baseip.format(random.randint(5, 250))


def write_samples(writer, dt, samples):
    for sample in samples:
        ts = calendar.timegm(dt.timetuple())
        sample['begin'] = ts
        sample['end'] = ts + 3600
        writer.writerow(sample)


def main():
    # Generators
    compute_var = ComputeVarMapper()
    image = ImageGenerator(3)
    image.generate_resources()
    volume = VolumeGenerator(2)
    volume.generate_resources()
    floating = FloatingGenerator(4, compute_var)
    floating.generate_resources()
    compute = ComputeGenerator(4, compute_var)
    compute.images = [resource['resource_id']
                      for resource in image.resources]
    compute.generate_resources()
    net_bw = NetworkBWGenerator(4, compute_var, NetBWVolMapper())
    net_bw.instances = [resource['resource_id']
                        for resource in compute.resources]
    net_bw.generate_resources()
    generators = [compute, image, volume, net_bw, floating]

    # Date
    now = datetime.datetime.utcnow()
    hour_delta = datetime.timedelta(hours=1)
    cur_date = now.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    cur_month = cur_date.month

    filename = sys.argv[1] if len(sys.argv) > 1 else 'generated.csv'
    with open(filename, 'wb') as csvfile:
        writer = csv.DictWriter(csvfile,
                                ['begin', 'end', 'type', 'desc', 'vol'])
        writer.writeheader()
        while cur_date.month == cur_month:
            for generator in generators:
                samples = generator.generate_samples(cur_date)
                write_samples(writer, cur_date, samples)
            cur_date += hour_delta


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

from oslo_config import cfg
from oslo_db import api as db_api
import six

_BACKEND_MAPPING = {
    'sqlalchemy': 'cloudkitty.rating.pyscripts.db.sqlalchemy.api'}
IMPL = db_api.DBAPI.from_config(cfg.CONF,
                                backend_mapping=_BACKEND_MAPPING,
                                lazy=True)


def get_instance():
    """Return a DB API instance."""
    return IMPL


class NoSuchScript(Exception):
    """Raised when the script doesn't exist."""

    def __init__(self, name=None, uuid=None):
        super(NoSuchScript, self).__init__(
            "No such script: %s (UUID: %s)" % (name, uuid))
        self.name = name
        self.uuid = uuid


class ScriptAlreadyExists(Exception):
    """Raised when the script already exists."""

    def __init__(self, name, uuid):
        super(ScriptAlreadyExists, self).__init__(
            "Script %s already exists (UUID: %s)" % (name, uuid))
        self.name = name
        self.uuid = uuid


@six.add_metaclass(abc.ABCMeta)
class PyScripts(object):
    """Base class for pyscripts configuration."""

    @abc.abstractmethod
    def get_migration(self):
        """Return a migrate manager.

        """

    @abc.abstractmethod
    def get_script(self, name=None, uuid=None):
        """Return a script object.

        :param name: Filter on a script name.
        :param uuid: The uuid of the script to get.
        """

    @abc.abstractmethod
    def list_scripts(self):
        """Return a UUID list of every scripts available.

        """

    @abc.abstractmethod
    def create_script(self, name, data):
        """Create a new script.

        :param name: Name of the script to create.
        :param data: Content of the python script.
        """

    @abc.abstractmethod
    def update_script(self, uuid, **kwargs):
        """Update a script.

        :param uuid UUID of the script to modify.
        :param data: Script data.
        """

    @abc.abstractmethod
    def delete_script(self, name=None, uuid=None):
        """Delete a list.

        :param name: Name of the script to delete.
        :param uuid: UUID of the script to delete.
        """



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_db import exception
from oslo_db.sqlalchemy import utils
from oslo_utils import uuidutils
import sqlalchemy

from cloudkitty import db
from cloudkitty.rating.pyscripts.db import api
from cloudkitty.rating.pyscripts.db.sqlalchemy import migration
from cloudkitty.rating.pyscripts.db.sqlalchemy import models


def get_backend():
    return PyScripts()


class PyScripts(api.PyScripts):

    def get_migration(self):
        return migration

    def get_script(self, name=None, uuid=None):
        session = db.get_session()
        try:
            q = session.query(models.PyScriptsScript)
            if name:
                q = q.filter(
                    models.PyScriptsScript.name == name)
            elif uuid:
                q = q.filter(
                    models.PyScriptsScript.script_id == uuid)
            else:
                raise ValueError('You must specify either name or uuid.')
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchScript(name=name, uuid=uuid)

    def list_scripts(self):
        session = db.get_session()
        q = session.query(models.PyScriptsScript)
        res = q.values(
            models.PyScriptsScript.script_id)
        return [uuid[0] for uuid in res]

    def create_script(self, name, data):
        session = db.get_session()
        try:
            with session.begin():
                script_db = models.PyScriptsScript(name=name)
                script_db.data = data
                script_db.script_id = uuidutils.generate_uuid()
                session.add(script_db)
            return script_db
        except exception.DBDuplicateEntry:
            script_db = self.get_script(name=name)
            raise api.ScriptAlreadyExists(
                script_db.name,
                script_db.script_id)

    def update_script(self, uuid, **kwargs):
        session = db.get_session()
        try:
            with session.begin():
                q = session.query(models.PyScriptsScript)
                q = q.filter(
                    models.PyScriptsScript.script_id == uuid
                )
                script_db = q.with_lockmode('update').one()
                if kwargs:
                    excluded_cols = ['script_id']
                    for col in excluded_cols:
                        if col in kwargs:
                            kwargs.pop(col)
                    for attribute, value in kwargs.items():
                        if hasattr(script_db, attribute):
                            setattr(script_db, attribute, value)
                        else:
                            raise ValueError('No such attribute: {}'.format(
                                attribute))
                else:
                    raise ValueError('No attribute to update.')
                return script_db
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchScript(uuid=uuid)

    def delete_script(self, name=None, uuid=None):
        session = db.get_session()
        q = utils.model_query(
            models.PyScriptsScript,
            session)
        if name:
            q = q.filter(models.PyScriptsScript.name == name)
        elif uuid:
            q = q.filter(models.PyScriptsScript.script_id == uuid)
        else:
            raise ValueError('You must specify either name or uuid.')
        r = q.delete()
        if not r:
            raise api.NoSuchScript(uuid=uuid)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import pecan
import six
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1 import types as ck_types
from cloudkitty import rating
from cloudkitty.rating.hash.datamodels import group as group_models
from cloudkitty.rating.hash.datamodels import threshold as threshold_models
from cloudkitty.rating.hash.db import api as db_api


class HashMapThresholdsController(rating.RatingRestControllerBase):
    """Controller responsible of thresholds management.

    """

    _custom_actions = {
        'group': ['GET']}

    @wsme_pecan.wsexpose(group_models.Group,
                         ck_types.UuidType())
    def group(self, threshold_id):
        """Get the group attached to the threshold.

        :param threshold_id: UUID of the threshold to filter on.
        """
        hashmap = db_api.get_instance()
        try:
            group_db = hashmap.get_group_from_threshold(
                uuid=threshold_id)
            return group_models.Group(**group_db.export_model())
        except db_api.ThresholdHasNoGroup as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(threshold_models.ThresholdCollection,
                         ck_types.UuidType(),
                         ck_types.UuidType(),
                         ck_types.UuidType(),
                         bool,
                         ck_types.UuidType(),
                         bool,
                         status_code=200)
    def get_all(self,
                service_id=None,
                field_id=None,
                group_id=None,
                no_group=False,
                tenant_id=None,
                filter_tenant=False):
        """Get the threshold list

        :param service_id: Service UUID to filter on.
        :param field_id: Field UUID to filter on.
        :param group_id: Group UUID to filter on.
        :param no_group: Filter on orphaned thresholds.
        :param tenant_id: Tenant UUID to filter on.
        :param filter_tenant: Explicitly filter on tenant (default is to not
        filter on tenant). Useful if you want to filter on tenant being None.
        :return: List of every thresholds.
        """
        hashmap = db_api.get_instance()
        threshold_list = []
        search_opts = dict()
        if filter_tenant:
            search_opts['tenant_uuid'] = tenant_id
        thresholds_uuid_list = hashmap.list_thresholds(
            service_uuid=service_id,
            field_uuid=field_id,
            group_uuid=group_id,
            **search_opts)
        for threshold_uuid in thresholds_uuid_list:
            threshold_db = hashmap.get_threshold(uuid=threshold_uuid)
            threshold_list.append(threshold_models.Threshold(
                **threshold_db.export_model()))
        res = threshold_models.ThresholdCollection(thresholds=threshold_list)
        return res

    @wsme_pecan.wsexpose(threshold_models.Threshold,
                         ck_types.UuidType())
    def get_one(self, threshold_id):
        """Return a threshold.

        :param threshold_id: UUID of the threshold to filter on.
        """
        hashmap = db_api.get_instance()
        try:
            threshold_db = hashmap.get_threshold(uuid=threshold_id)
            return threshold_models.Threshold(
                **threshold_db.export_model())
        except db_api.NoSuchThreshold as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(threshold_models.Threshold,
                         body=threshold_models.Threshold,
                         status_code=201)
    def post(self, threshold_data):
        """Create a threshold.

        :param threshold_data: Informations about the threshold to create.
        """
        hashmap = db_api.get_instance()
        try:
            threshold_db = hashmap.create_threshold(
                level=threshold_data.level,
                map_type=threshold_data.map_type,
                cost=threshold_data.cost,
                field_id=threshold_data.field_id,
                group_id=threshold_data.group_id,
                service_id=threshold_data.service_id,
                tenant_id=threshold_data.tenant_id)
            pecan.response.location = pecan.request.path_url
            if pecan.response.location[-1] != '/':
                pecan.response.location += '/'
            pecan.response.location += threshold_db.threshold_id
            return threshold_models.Threshold(
                **threshold_db.export_model())
        except db_api.ThresholdAlreadyExists as e:
            pecan.abort(409, six.text_type(e))
        except db_api.ClientHashMapError as e:
            pecan.abort(400, six.text_type(e))

    @wsme_pecan.wsexpose(None,
                         ck_types.UuidType(),
                         body=threshold_models.Threshold,
                         status_code=302)
    def put(self, threshold_id, threshold):
        """Update a threshold.

        :param threshold_id: UUID of the threshold to update.
        :param threshold: Threshold data to insert.
        """
        hashmap = db_api.get_instance()
        try:
            hashmap.update_threshold(
                threshold_id,
                threshold_id=threshold.threshold_id,
                level=threshold.level,
                cost=threshold.cost,
                map_type=threshold.map_type,
                group_id=threshold.group_id,
                tenant_id=threshold.tenant_id)
            pecan.response.headers['Location'] = pecan.request.path
        except db_api.NoSuchThreshold as e:
            pecan.abort(404, six.text_type(e))
        except db_api.ClientHashMapError as e:
            pecan.abort(400, six.text_type(e))

    @wsme_pecan.wsexpose(None,
                         ck_types.UuidType(),
                         status_code=204)
    def delete(self, threshold_id):
        """Delete a threshold.

        :param threshold_id: UUID of the threshold to delete.
        """
        hashmap = db_api.get_instance()
        try:
            hashmap.delete_threshold(uuid=threshold_id)
        except db_api.NoSuchThreshold as e:
            pecan.abort(404, six.text_type(e))



# Copyright 2017 GohighSec.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

from cloudkitty.common.policies import base
from cloudkitty.common.policies import collector
from cloudkitty.common.policies import info
from cloudkitty.common.policies import rating
from cloudkitty.common.policies import report
from cloudkitty.common.policies import storage


def list_rules():
    return itertools.chain(
        base.list_rules(),
        collector.list_rules(),
        info.list_rules(),
        rating.list_rules(),
        report.list_rules(),
        storage.list_rules()
    )



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import os

from gabbi import driver

from cloudkitty.tests.gabbi import fixtures
from cloudkitty.tests.gabbi.rating.pyscripts import fixtures as py_fixtures

TESTS_DIR = 'gabbits'


def load_tests(loader, tests, pattern):
    test_dir = os.path.join(os.path.dirname(__file__), TESTS_DIR)
    return driver.build_tests(test_dir,
                              loader,
                              host=None,
                              intercept=fixtures.setup_app,
                              fixture_module=py_fixtures)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

from oslo_config import cfg
from oslo_log import log as logging
import six
from stevedore import driver

from cloudkitty import collector as ck_collector
from cloudkitty import utils as ck_utils

LOG = logging.getLogger(__name__)

storage_opts = [
    cfg.StrOpt('backend',
               default='sqlalchemy',
               help='Name of the storage backend driver.')
]

CONF = cfg.CONF
CONF.register_opts(storage_opts, group='storage')

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

STORAGES_NAMESPACE = 'cloudkitty.storage.backends'


def get_storage(collector=None):
    storage_args = {
        'period': METRICS_CONF['period'],
        'collector': collector if collector else ck_collector.get_collector()}
    backend = driver.DriverManager(
        STORAGES_NAMESPACE,
        cfg.CONF.storage.backend,
        invoke_on_load=True,
        invoke_kwds=storage_args).driver
    if cfg.CONF.storage.backend not in ['sqlalchemy', 'hybrid']:
        LOG.warning('{} storage backend is deprecated and will be removed '
                    'in a future release.'.format(cfg.CONF.storage.backend))
    return backend


class NoTimeFrame(Exception):
    """Raised when there is no time frame available."""

    def __init__(self):
        super(NoTimeFrame, self).__init__(
            "No time frame available")


@six.add_metaclass(abc.ABCMeta)
class BaseStorage(object):
    """Base Storage class:

        Handle incoming data from the global orchestrator, and store them.
    """
    def __init__(self, **kwargs):
        self._period = kwargs.get('period', METRICS_CONF['period'])
        self._collector = kwargs.get('collector')

        # State vars
        self.usage_start = {}
        self.usage_start_dt = {}
        self.usage_end = {}
        self.usage_end_dt = {}
        self._has_data = {}

    @staticmethod
    def init():
        """Initialize storage backend.

        Can be used to create DB schema on first start.
        """
        pass

    def _filter_period(self, json_data):
        """Detect the best usage period to extract.

        Removes the usage from the json data and returns it.
        :param json_data: Data to filter.
        """
        candidate_ts = None
        candidate_idx = 0

        for idx, usage in enumerate(json_data):
            usage_ts = usage['period']['begin']
            if candidate_ts is None or usage_ts < candidate_ts:
                candidate_ts = usage_ts
                candidate_idx = idx

        if candidate_ts:
            return candidate_ts, json_data.pop(candidate_idx)['usage']

    def _pre_commit(self, tenant_id):
        """Called before every commit.

        :param tenant_id: tenant_id which information must be committed.
        """

    @abc.abstractmethod
    def _commit(self, tenant_id):
        """Push data to the storage backend.

        :param tenant_id: tenant_id which information must be committed.
        """

    def _post_commit(self, tenant_id):
        """Called after every commit.

        :param tenant_id: tenant_id which information must be committed.
        """
        if tenant_id in self._has_data:
            del self._has_data[tenant_id]
        self._clear_usage_info(tenant_id)

    @abc.abstractmethod
    def _dispatch(self, data, tenant_id):
        """Process rated data.

        :param data: The rated data frames.
        :param tenant_id: tenant_id which data must be dispatched to.
        """

    def _update_start(self, begin, tenant_id):
        """Update usage_start with a new timestamp.

        :param begin: New usage beginning timestamp.
        :param tenant_id: tenant_id to update.
        """
        self.usage_start[tenant_id] = begin
        self.usage_start_dt[tenant_id] = ck_utils.ts2dt(begin)

    def _update_end(self, end, tenant_id):
        """Update usage_end with a new timestamp.

        :param end: New usage end timestamp.
        :param tenant_id: tenant_id to update.
        """
        self.usage_end[tenant_id] = end
        self.usage_end_dt[tenant_id] = ck_utils.ts2dt(end)

    def _clear_usage_info(self, tenant_id):
        """Clear usage information timestamps.

        :param tenant_id: tenant_id which information needs to be removed.
        """
        self.usage_start.pop(tenant_id, None)
        self.usage_start_dt.pop(tenant_id, None)
        self.usage_end.pop(tenant_id, None)
        self.usage_end_dt.pop(tenant_id, None)

    def _check_commit(self, usage_start, tenant_id):
        """Check if the period for a given tenant must be committed.

        :param usage_start: Start of the period.
        :param tenant_id: tenant_id to check for.
        """
        usage_end = self.usage_end.get(tenant_id)
        if usage_end is not None and usage_start >= usage_end:
            self.commit(tenant_id)
        if self.usage_start.get(tenant_id) is None:
            self._update_start(usage_start, tenant_id)
            self._update_end(usage_start + self._period, tenant_id)

    @abc.abstractmethod
    def get_state(self, tenant_id=None):
        """Return the last written frame's timestamp.

        :param tenant_id: tenant_id to filter on.
        """

    @abc.abstractmethod
    def get_total(self, begin=None, end=None, tenant_id=None,
                  service=None, groupby=None):
        """Return the current total.

        :param begin: When to start filtering.
        :type begin: datetime.datetime
        :param end: When to stop filtering.
        :type end: datetime.datetime
        :param tenant_id: Filter on the tenant_id.
        :type res_type: str
        :param service: Filter on the resource type.
        :type service: str
        :param groupby: Fields to group by, separated by commas if multiple.
        :type groupby: str
        """

    @abc.abstractmethod
    def get_tenants(self, begin, end):
        """Return the list of rated tenants.

        :param begin: When to start filtering.
        :type begin: datetime.datetime
        :param end: When to stop filtering.
        :type end: datetime.datetime
        """

    @abc.abstractmethod
    def get_time_frame(self, begin, end, **filters):
        """Request a time frame from the storage backend.

        :param begin: When to start filtering.
        :type begin: datetime.datetime
        :param end: When to stop filtering.
        :type end: datetime.datetime
        :param res_type: (Optional) Filter on the resource type.
        :type res_type: str
        :param tenant_id: (Optional) Filter on the tenant_id.
        :type res_type: str
        """

    def append(self, raw_data, tenant_id):
        """Append rated data before committing them to the backend.

        :param raw_data: The rated data frames.
        :param tenant_id: Tenant the frame is belonging to.
        """
        while raw_data:
            usage_start, data = self._filter_period(raw_data)
            self._check_commit(usage_start, tenant_id)
            self._dispatch(data, tenant_id)

    def nodata(self, begin, end, tenant_id):
        """Append a no data frame to the storage backend.

        :param begin: Begin of the period with no data.
        :param end: End of the period with no data.
        :param tenant_id: Tenant to update with no data marker for the period.
        """
        self._check_commit(begin, tenant_id)

    def commit(self, tenant_id):
        """Commit the changes to the backend.

        :param tenant_id: Tenant the changes belong to.
        """
        self._pre_commit(tenant_id)
        self._commit(tenant_id)
        self._post_commit(tenant_id)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal

from wsme import types as wtypes

from cloudkitty.api.v1 import types as ck_types
from cloudkitty.rating.hash.datamodels import mapping as mapping_models


class Threshold(wtypes.Base):
    """Type describing a Threshold.

    A threshold is used to apply rating rules based on a level, if the parent
    is a field then the level is checked against a metadata. If it's a service
    then it's the quantity of the resource that is checked.
    """

    threshold_id = wtypes.wsattr(ck_types.UuidType(), mandatory=False)
    """UUID of the threshold."""

    level = wtypes.wsattr(decimal.Decimal,
                          mandatory=True,
                          default=decimal.Decimal('0'))
    """Level of the threshold."""

    map_type = wtypes.wsattr(mapping_models.MAP_TYPE,
                             default='flat',
                             name='type')
    """Type of the threshold."""

    cost = wtypes.wsattr(decimal.Decimal, mandatory=True)
    """Value of the threshold."""

    service_id = wtypes.wsattr(ck_types.UuidType(),
                               mandatory=False)
    """UUID of the service."""

    field_id = wtypes.wsattr(ck_types.UuidType(),
                             mandatory=False)
    """UUID of the field."""

    group_id = wtypes.wsattr(ck_types.UuidType(),
                             mandatory=False)
    """UUID of the hashmap group."""

    tenant_id = wtypes.wsattr(ck_types.UuidType(),
                              mandatory=False,
                              default=None)
    """UUID of the hashmap tenant."""

    @classmethod
    def sample(cls):
        sample = cls(threshold_id='39dbd39d-f663-4444-a795-fb19d81af136',
                     field_id='ac55b000-a05b-4832-b2ff-265a034886ab',
                     level=decimal.Decimal('1024'),
                     map_type='flat',
                     cost=decimal.Decimal('4.2'),
                     tenant_id='7977999e-2e25-11e6-a8b2-df30b233ffcb')
        return sample


class ThresholdCollection(wtypes.Base):
    """Type describing a list of mappings.

    """

    thresholds = [Threshold]
    """List of thresholds."""

    @classmethod
    def sample(cls):
        sample = Threshold.sample()
        return cls(thresholds=[sample])



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_utils import uuidutils
from wsme import types as wtypes


class UuidType(wtypes.UuidType):
    """A simple UUID type."""
    basetype = wtypes.text
    name = 'uuid'

    @staticmethod
    def validate(value):
        if not uuidutils.is_uuid_like(value):
            raise ValueError("Invalid UUID, got '%s'" % value)
        return value


# Code taken from ironic types
class MultiType(wtypes.UserType):
    """A complex type that represents one or more types.

    Used for validating that a value is an instance of one of the types.

    :param *types: Variable-length list of types.

    """
    def __init__(self, *types):
        self.types = types

    def __str__(self):
        return ' | '.join(map(str, self.types))

    def validate(self, value):
        for t in self.types:
            if t is wtypes.text and isinstance(value, wtypes.bytes):
                value = value.decode()
            if isinstance(value, t):
                return value
        else:
            raise ValueError(
                "Wrong type. Expected '%(type)s', got '%(value)s'"
                % {'type': self.types, 'value': type(value)})



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import collections
import csv
import datetime
import os

from cloudkitty import utils as ck_utils
from cloudkitty import writer


class InconsistentHeaders(Exception):
    pass


class BaseCSVBackend(writer.BaseReportWriter):
    """Report format writer:

        Generates report in csv format
    """
    report_type = 'csv'

    def __init__(self, write_orchestrator, user_id, backend, basepath):
        super(BaseCSVBackend, self).__init__(write_orchestrator,
                                             user_id,
                                             backend,
                                             basepath)

        # Detailed transform OrderedDict
        self._field_map = collections.OrderedDict()

        self._headers = []
        self._headers_len = 0
        self._extra_headers = []
        self._extra_headers_len = 0

        # File vars
        self._csv_report = None

        # State variables
        self.cached_start = None
        self.cached_start_str = ''
        self.cached_end = None
        self.cached_end_str = ''
        self._crumpled = False

        # Current usage period lines
        self._usage_data = []

    def _gen_filename(self, timeframe):
        filename = ('{}-{}-{:02d}.csv').format(self._tenant_id,
                                               timeframe.year,
                                               timeframe.month)
        if self._basepath:
            filename = os.path.join(self._basepath, filename)
        return filename

    def _open(self):
        filename = self._gen_filename(self.usage_start_dt)
        self._report = self._backend(filename, 'rb+')
        self._csv_report = csv.writer(self._report)
        self._report.seek(0, 2)

    def _close_file(self):
        if self._report is not None:
            self._report.close()

    def _get_state_manager_timeframe(self):
        if self.report_type is None:
            raise NotImplementedError()

    def _update_state_manager(self):
        if self.report_type is None:
            raise NotImplementedError()

        super(BaseCSVBackend, self)._update_state_manager()

        metadata = {'total': self.total}
        metadata['headers'] = self._extra_headers
        self._sm.set_metadata(metadata)

    def _init_headers(self):
        headers = self._field_map.keys()
        for header in headers:
            if ':*' in header:
                continue
            self._headers.append(header)
        self._headers_len = len(self._headers)

    def _write_header(self):
        self._csv_report.writerow(self._headers + self._extra_headers)

    def _write(self):
        self._csv_report.writerows(self._usage_data)

    def _post_commit(self):
        self._crumpled = False
        self._usage_data = []
        self._write_total()

    def _update(self, data):
        """Dispatch report data with context awareness.

        """
        if self._crumpled:
            return
        try:
            for service in data:
                for report_data in data[service]:
                    self._process_data(service, report_data)
                    self.total += report_data['rating']['price']
        except InconsistentHeaders:
            self._crumple()
            self._crumpled = True

    def _recover_state(self):
        # Rewind 3 lines
        self._report.seek(0, 2)
        buf_size = self._report.tell()
        if buf_size > 2000:
            buf_size = 2000
        elif buf_size == 0:
            return
        self._report.seek(-buf_size, 2)
        end_buf = self._report.read()
        last_line = buf_size
        for dummy in range(4):
            last_line = end_buf.rfind('\n', 0, last_line)
        if last_line > 0:
            last_line -= len(end_buf) - 1
        else:
            raise RuntimeError('Unable to recover file state.')
        self._report.seek(last_line, 2)
        self._report.truncate()

    def _crumple(self):
        # Reset states
        self._usage_data = []
        self.total = 0

        # Recover state from file
        if self._report is not None:
            self._report.seek(0)
            reader = csv.reader(self._report)
            # Skip header
            for dummy in range(2):
                line = reader.next()
            self.usage_start_dt = datetime.datetime.strptime(
                line[0],
                '%Y/%m/%d %H:%M:%S')
            self.usage_start = ck_utils.dt2ts(self.usage_start_dt)
            self.usage_end_dt = datetime.datetime.strptime(
                line[1],
                '%Y/%m/%d %H:%M:%S')
            self.usage_end = ck_utils.dt2ts(self.usage_end_dt)

            # Reset file
            self._report.seek(0)
            self._report.truncate()
            self._write_header()

        timeframe = self._write_orchestrator.get_timeframe(
            self.usage_start)
        start = self.usage_start
        self.usage_start = None
        for data in timeframe:
            self.append(data['usage'],
                        start,
                        None)
        self.usage_start = self.usage_end

    def _update_extra_headers(self, new_head):
        self._extra_headers.append(new_head)
        self._extra_headers.sort()
        self._extra_headers_len += 1

    def _allocate_extra(self, line):
        for dummy in range(self._extra_headers_len):
            line.append('')

    def _map_wildcard(self, base, report_data):
        wildcard_line = []
        headers_changed = False
        self._allocate_extra(wildcard_line)
        base_section, dummy = base.split(':')
        if not report_data:
            return []
        for field in report_data:
            col_name = base_section + ':' + field
            if col_name not in self._extra_headers:
                self._update_extra_headers(col_name)
                headers_changed = True
            else:
                idx = self._extra_headers.index(col_name)
                wildcard_line[idx] = report_data[field]
        if headers_changed:
            raise InconsistentHeaders('Headers value changed'
                                      ', need to rebuild.')
        return wildcard_line

    def _recurse_sections(self, sections, data):
        if not sections.count(':'):
            return data.get(sections, '')
        fields = sections.split(':')
        cur_data = data
        for field in fields:
            if field in cur_data:
                cur_data = cur_data[field]
            else:
                return None
        return cur_data

    def _process_data(self, context, report_data):
        """Transform the raw json data to the final CSV values.

        """
        if not self._headers_len:
            self._init_headers()

        formated_data = []
        for base, mapped in self._field_map.iteritems():
            final_data = ''
            if isinstance(mapped, str):
                mapped_section, mapped_field = mapped.rsplit(':', 1)
                data = self._recurse_sections(mapped_section, report_data)
                if mapped_field == '*':
                    extra_fields = self._map_wildcard(base, data)
                    formated_data.extend(extra_fields)
                    continue
                elif mapped_section in report_data:
                    data = report_data[mapped_section]
                    if mapped_field in data:
                        final_data = data[mapped_field]
            elif mapped is not None:
                final_data = mapped(context, report_data)
            formated_data.append(final_data)

        self._usage_data.append(formated_data)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_concurrency import lockutils
import pecan
from pecan import rest
from stevedore import extension
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1.datamodels import rating as rating_models
from cloudkitty.common import policy
from cloudkitty import utils as ck_utils

PROCESSORS_NAMESPACE = 'cloudkitty.rating.processors'


class RatingModulesMixin(object):
    def reload_extensions(self):
        lock = lockutils.lock('rating-modules')
        with lock:
            ck_utils.refresh_stevedore(PROCESSORS_NAMESPACE)
            # FIXME(sheeprine): Implement RPC messages to trigger reload on
            # processors
            self.extensions = extension.ExtensionManager(
                PROCESSORS_NAMESPACE,
                # FIXME(sheeprine): don't want to load it here as we just need
                # the controller
                invoke_on_load=True)
            if not self._first_call:
                self.notify_reload()
            else:
                self._first_call = False

    def notify_reload(self):
        client = pecan.request.rpc_client.prepare(namespace='rating',
                                                  version='1.1')
        client.cast({}, 'reload_modules')

    def __init__(self):
        self._first_call = True
        self.extensions = []
        self.reload_extensions()


class ModulesController(rest.RestController, RatingModulesMixin):
    """REST Controller managing rating modules."""

    def route(self, *args):
        route = args[0]
        if route.startswith('/v1/module_config'):
            policy.authorize(pecan.request.context, 'rating:module_config', {})

        super(ModulesController, self).route(*args)

    @wsme_pecan.wsexpose(rating_models.CloudkittyModuleCollection)
    def get_all(self):
        """return the list of loaded modules.

        :return: name of every loaded modules.
        """
        policy.authorize(pecan.request.context, 'rating:list_modules', {})

        modules_list = []
        lock = lockutils.lock('rating-modules')
        with lock:
            for module in self.extensions:
                infos = module.obj.module_info.copy()
                infos['module_id'] = infos.pop('name')
                modules_list.append(rating_models.CloudkittyModule(**infos))

        return rating_models.CloudkittyModuleCollection(
            modules=modules_list)

    @wsme_pecan.wsexpose(rating_models.CloudkittyModule, wtypes.text)
    def get_one(self, module_id):
        """return a module

        :return: CloudKittyModule
        """
        policy.authorize(pecan.request.context, 'rating:get_module', {})

        try:
            lock = lockutils.lock('rating-modules')
            with lock:
                module = self.extensions[module_id]
        except KeyError:
            pecan.abort(404, 'Module not found.')
        infos = module.obj.module_info.copy()
        infos['module_id'] = infos.pop('name')
        return rating_models.CloudkittyModule(**infos)

    @wsme_pecan.wsexpose(rating_models.CloudkittyModule,
                         wtypes.text,
                         body=rating_models.CloudkittyModule,
                         status_code=302)
    def put(self, module_id, module):
        """Change the state and priority of a module.

        :param module_id: name of the module to modify
        :param module: CloudKittyModule object describing the new desired state
        """
        policy.authorize(pecan.request.context, 'rating:update_module', {})

        try:
            lock = lockutils.lock('rating-modules')
            with lock:
                ext = self.extensions[module_id].obj
        except KeyError:
            pecan.abort(404, 'Module not found.')
        if module.enabled != wtypes.Unset and ext.enabled != module.enabled:
            ext.set_state(module.enabled)
        if module.priority != wtypes.Unset and ext.priority != module.priority:
            ext.set_priority(module.priority)
        pecan.response.location = pecan.request.path


class UnconfigurableController(rest.RestController):
    """This controller raises an error when requested."""

    @wsme_pecan.wsexpose(None)
    def _default(self):
        self.abort()

    def abort(self):
        pecan.abort(409, "Module is not configurable")


class ModulesExposer(rest.RestController, RatingModulesMixin):
    """REST Controller exposing rating modules.

    This is the controller that exposes the modules own configuration
    settings.
    """

    def __init__(self):
        super(ModulesExposer, self).__init__()
        self._loaded_modules = []
        self.expose_modules()

    def expose_modules(self):
        """Load rating modules to expose API controllers."""
        lock = lockutils.lock('rating-modules')
        with lock:
            for ext in self.extensions:
                # FIXME(sheeprine): we should notify two modules with same name
                name = ext.name
                if not ext.obj.config_controller:
                    ext.obj.config_controller = UnconfigurableController
                # Update extension reference
                setattr(self, name, ext.obj.config_controller())
                if name in self._loaded_modules:
                    self._loaded_modules.remove(name)
            # Clear removed modules
            for module in self._loaded_modules:
                delattr(self, module)
            self._loaded_modules = self.extensions.names()


class RatingController(rest.RestController):
    """The RatingController is exposed by the API.

    The RatingControler connects the ModulesExposer, ModulesController
    and a quote action to the API.
    """

    _custom_actions = {
        'quote': ['POST'],
        'reload_modules': ['GET'],
    }

    modules = ModulesController()
    module_config = ModulesExposer()

    @wsme_pecan.wsexpose(float,
                         body=rating_models.CloudkittyResourceCollection)
    def quote(self, res_data):
        """Get an instant quote based on multiple resource descriptions.

        :param res_data: List of resource descriptions.
        :return: Total price for these descriptions.
        """
        policy.authorize(pecan.request.context, 'rating:quote', {})

        client = pecan.request.rpc_client.prepare(namespace='rating')
        res_dict = {}
        for res in res_data.resources:
            if res.service not in res_dict:
                res_dict[res.service] = []
            json_data = res.to_json()
            res_dict[res.service].extend(json_data[res.service])

        res = client.call({}, 'quote', res_data=[{'usage': res_dict}])
        return res

    @wsme_pecan.wsexpose(None)
    def reload_modules(self):
        """Trigger a rating module list reload.

        """
        policy.authorize(pecan.request.context, 'rating:module_config', {})
        self.modules.reload_extensions()
        self.module_config.reload_extensions()
        self.module_config.expose_modules()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_config import cfg
from oslo_db import options as db_options  # noqa
from oslo_messaging import opts  # noqa


state_opts = [
    cfg.StrOpt('backend',
               default='cloudkitty.backend.file.FileBackend',
               help='Backend for the state manager.'),
    cfg.StrOpt('basepath',
               default='/var/lib/cloudkitty/states/',
               help='Storage directory for the file state backend.'), ]

output_opts = [
    cfg.StrOpt('backend',
               default='cloudkitty.backend.file.FileBackend',
               help='Backend for the output manager.'),
    cfg.StrOpt('basepath',
               default='/var/lib/cloudkitty/states/',
               help='Storage directory for the file output backend.'),
    cfg.ListOpt('pipeline',
                default=['osrf'],
                help='Output pipeline'), ]


cfg.CONF.register_opts(state_opts, 'state')
cfg.CONF.register_opts(output_opts, 'output')

# oslo.db defaults
db_options.set_defaults(
    cfg.CONF,
    connection='sqlite:////var/lib/cloudkitty/cloudkitty.sqlite')



# -*- coding: utf-8 -*-
# Copyright 2016 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import os

from gabbi.handlers import base


class EnvironStoreHandler(base.ResponseHandler):
    """Hackish response handler used to store data in environment.

    Store data into an environment variable to implement some kind of variable
    use in gabbi.
    """
    test_key_suffix = 'store_environ'
    test_key_value = {}

    def action(self, test, key, value=None):
        data = test.content_handlers[0].extract_json_path_value(
            test.response_data, value)
        os.environ[key] = test.replace_template(data)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import collections
import datetime

from cloudkitty.writer import csv_base


class CSVMapped(csv_base.BaseCSVBackend):
    report_type = 'csv'

    def __init__(self, write_orchestrator, user_id, backend, state_backend):
        super(CSVMapped, self).__init__(write_orchestrator,
                                        user_id,
                                        backend,
                                        state_backend)

        # Detailed transform dict
        self._field_map = collections.OrderedDict(
            [('UsageStart', self._trans_get_usage_start),
             ('UsageEnd', self._trans_get_usage_end),
             ('ResourceId', self._trans_res_id),
             ('Operation', self._trans_operation),
             ('UserId', 'desc:user_id'),
             ('ProjectId', 'desc:project_id'),
             ('ItemName', 'desc:name'),
             ('ItemFlavor', 'desc:flavor_name'),
             ('ItemFlavorId', 'desc:flavor_id'),
             ('AvailabilityZone', 'desc:availability_zone'),
             ('Service', self._trans_service),
             ('UsageQuantity', 'vol:qty'),
             ('RateValue', 'rating:price'),
             ('Cost', self._trans_calc_cost),
             ('user:*', 'desc:metadata:*')])

    def _write_total(self):
        lines = [[''] * self._headers_len for i in range(3)]
        for i in range(len(lines)):
            lines[i][1] = self._tenant_id

        lines[1][2] = self._tenant_id

        lines[0][3] = 'InvoiceTotal'
        lines[1][3] = 'AccountTotal'
        lines[2][3] = 'StatementTotal'

        lines[0][5] = 'Total amount for invoice'
        lines[1][5] = 'Total for linked account# {}'.format(self._tenant_id)
        start_month = datetime.datetime(
            self.usage_start_dt.year,
            self.usage_start_dt.month,
            1)
        lines[2][5] = ('Total statement amount for period '
                       '{} - {}').format(self._format_date(start_month),
                                         self._get_usage_end())

        lines[0][8] = self.total
        lines[1][8] = self.total
        lines[2][8] = self.total

        self._csv_report.writerows(lines)

    @staticmethod
    def _format_date(raw_dt):
        return raw_dt.strftime('%Y/%m/%d %H:%M:%S')

    def _get_usage_start(self):
        """Get the start usage of this period.

        """
        if self.cached_start == self.usage_start:
            return self.cached_start_str
        else:
            self.cached_start = self.usage_start
            self.cached_start_str = self._format_date(self.usage_start_dt)
            return self.cached_start_str

    def _get_usage_end(self):
        """Get the end usage of this period.

        """
        if self.cached_start == self.usage_start and self.cached_end_str \
           and self.cached_end > self.cached_start:
            return self.cached_end_str
        else:
            usage_end = self.usage_start_dt + datetime.timedelta(
                seconds=self._period)
            self.cached_end_str = self._format_date(usage_end)
            return self.cached_end_str

    def _trans_get_usage_start(self, _context, _report_data):
        """Dummy transformation function to comply with the standard.

        """
        return self._get_usage_start()

    def _trans_get_usage_end(self, _context, _report_data):
        """Dummy transformation function to comply with the standard.

        """
        return self._get_usage_end()

    def _trans_product_name(self, context, _report_data):
        """Context dependent product name translation.

        """
        if context == 'compute':
            return 'Nova Computing'
        else:
            return context

    def _trans_operation(self, context, _report_data):
        """Context dependent operation translation.

        """
        if context == 'compute':
            return 'RunInstances'

    def _trans_res_id(self, context, report_data):
        """Context dependent resource id transformation function.

        """
        return report_data['desc'].get('resource_id')

    def _trans_calc_cost(self, context, report_data):
        """Cost calculation function.

        """
        try:
            quantity = report_data['vol'].get('qty')
            rate = report_data['rating'].get('price')
            return str(float(quantity) * rate)
        except TypeError:
            pass

    def _trans_service(self, context, report_data):
        return context



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Borrowed from cinder (cinder/policy.py)

import copy
import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import opts as policy_opts
from oslo_policy import policy
from oslo_utils import excutils
import six

from cloudkitty.common import policies

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
policy_opts.set_defaults(cfg.CONF, 'policy.json')

_ENFORCER = None
# oslo_policy will read the policy configuration file again when the file
# is changed in runtime so the old policy rules will be saved to
# saved_file_rules and used to compare with new rules to determine the
# rules whether were updated.
saved_file_rules = []


# TODO(gpocentek): provide a proper parent class to handle such exceptions
class PolicyNotAuthorized(Exception):
    message = "Policy doesn't allow %(action)s to be performed."
    code = 403

    def __init__(self, **kwargs):
        self.msg = self.message % kwargs
        super(PolicyNotAuthorized, self).__init__(self.msg)

    def __unicode__(self):
        return six.text_type(self.msg)


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
        _ENFORCER = None


def init():
    global _ENFORCER
    global saved_file_rules
    if not _ENFORCER:
        _ENFORCER = policy.Enforcer(CONF)
        register_rules(_ENFORCER)

    # Only the rules which are loaded from file may be changed.
    current_file_rules = _ENFORCER.file_rules
    current_file_rules = _serialize_rules(current_file_rules)

    # Checks whether the rules are updated in the runtime
    if saved_file_rules != current_file_rules:
        saved_file_rules = copy.deepcopy(current_file_rules)


def _serialize_rules(rules):
    """Serialize all the Rule object as string."""
    result = [(rule_name, str(rule))
              for rule_name, rule in rules.items()]
    return sorted(result, key=lambda rule: rule[0])


def authorize(context, action, target):
    """Verifies that the action is valid on the target in this context.

       :param context: cloudkitty context
       :param action: string representing the action to be checked
           this should be colon separated for clarity.
           i.e. ``compute:create_instance``,
           ``compute:attach_volume``,
           ``volume:attach_volume``

       :param object: dictionary representing the object of the action
           for object creation this should be a dictionary representing the
           location of the object e.g. ``{'project_id': context.project_id}``

       :raises PolicyNotAuthorized: if verification fails.

    """
    if CONF.auth_strategy != "keystone":
        return

    init()

    try:
        return _ENFORCER.authorize(action, target, context.to_dict(),
                                   do_raise=True,
                                   exc=PolicyNotAuthorized,
                                   action=action)

    except policy.PolicyNotRegistered:
        with excutils.save_and_reraise_exception():
            LOG.exception('Policy not registered')
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.error('Policy check for %(action)s failed with credentials '
                      '%(credentials)s',
                      {'action': action, 'credentials': context.to_dict()})


def check_is_admin(roles):
    """Whether or not roles contains 'admin' role according to policy setting.

    """
    if CONF.auth_strategy != "keystone":
        return True

    init()

    # include project_id on target to avoid KeyError if context_is_admin
    # policy definition is missing, and default admin_or_owner rule
    # attempts to apply.  Since our credentials dict does not include a
    # project_id, this target can never match as a generic rule.
    target = {'project_id': ''}
    credentials = {'roles': roles}

    return _ENFORCER.authorize('context_is_admin', target, credentials)


def register_rules(enforcer):
    enforcer.register_defaults(policies.list_rules())


def get_enforcer():
    # This method is for use by oslopolicy CLI scripts. Those scripts need the
    # 'output-file' and 'namespace' options, but having those in sys.argv means
    # loading the Cloudkitty config options will fail as those are not expected
    # to be present. So we pass in an arg list with those stripped out.
    conf_args = []
    # Start at 1 because cfg.CONF expects the equivalent of sys.argv[1:]
    i = 1
    while i < len(sys.argv):
        if sys.argv[i].strip('-') in ['namespace', 'output-file']:
            i += 2
            continue
        conf_args.append(sys.argv[i])
        i += 1

    cfg.CONF(conf_args, project='cloudkitty')
    init()
    return _ENFORCER



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal

from wsme import types as wtypes

from cloudkitty.api.v1 import types as ck_types

MAP_TYPE = wtypes.Enum(wtypes.text, 'flat', 'rate')


class Mapping(wtypes.Base):
    """Type describing a Mapping.

    A mapping is used to apply rating rules based on a value, if the parent is
    a field then it's check the value of a metadata. If it's a service then it
    directly apply the rate to the volume.
    """

    mapping_id = wtypes.wsattr(ck_types.UuidType(), mandatory=False)
    """UUID of the mapping."""

    value = wtypes.wsattr(wtypes.text, mandatory=False, default='')
    """Key of the mapping."""

    map_type = wtypes.wsattr(MAP_TYPE, default='flat', name='type')
    """Type of the mapping."""

    cost = wtypes.wsattr(decimal.Decimal, mandatory=True)
    """Value of the mapping."""

    service_id = wtypes.wsattr(ck_types.UuidType(),
                               mandatory=False)
    """UUID of the service."""

    field_id = wtypes.wsattr(ck_types.UuidType(),
                             mandatory=False)
    """UUID of the field."""

    group_id = wtypes.wsattr(ck_types.UuidType(),
                             mandatory=False)
    """UUID of the hashmap group."""

    tenant_id = wtypes.wsattr(ck_types.UuidType(),
                              mandatory=False,
                              default=None)
    """UUID of the hashmap tenant."""

    @classmethod
    def sample(cls):
        sample = cls(mapping_id='39dbd39d-f663-4444-a795-fb19d81af136',
                     field_id='ac55b000-a05b-4832-b2ff-265a034886ab',
                     value='m1.micro',
                     map_type='flat',
                     cost=decimal.Decimal('4.2'),
                     tenant_id='7977999e-2e25-11e6-a8b2-df30b233ffcb')
        return sample


class MappingCollection(wtypes.Base):
    """Type describing a list of mappings.

    """

    mappings = [Mapping]
    """List of mappings."""

    @classmethod
    def sample(cls):
        sample = Mapping.sample()
        return cls(mappings=[sample])



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
from cloudkitty.collector import gnocchi
from cloudkitty import tests
from cloudkitty.tests import samples


class GnocchiCollectorTest(tests.TestCase):
    def setUp(self):
        super(GnocchiCollectorTest, self).setUp()
        self._tenant_id = samples.TENANT
        self.collector = gnocchi.GnocchiCollector

    # Filter generation
    def test_generate_one_field_filter(self):
        actual = self.collector.gen_filter(value1=2)
        expected = {
            '=': {
                'value1': 2
            }}
        self.assertEqual(expected, actual)

    def test_generate_two_fields_filter(self):
        actual = self.collector.gen_filter(value1=2, value2=3)
        expected = {'and': [{
            '=': {
                'value1': 2
            }}, {
            '=': {
                'value2': 3
            }}]}
        self.assertEqual(expected, actual)

    def test_generate_two_fields_filter_different_operations(self):
        actual = self.collector.gen_filter(
            cop='>=',
            lop='or',
            value1=2,
            value2=3)
        expected = {'or': [{
            '>=': {
                'value1': 2
            }}, {
            '>=': {
                'value2': 3
            }}]}
        self.assertEqual(expected, actual)

    def test_generate_two_filters_and_add_logical(self):
        filter1 = self.collector.gen_filter(value1=2)
        filter2 = self.collector.gen_filter(cop='>', value2=3)
        actual = self.collector.extend_filter(filter1, filter2, lop='or')
        expected = {'or': [{
            '=': {
                'value1': 2
            }}, {
            '>': {
                'value2': 3
            }}]}
        self.assertEqual(expected, actual)

    def test_noop_on_single_filter(self):
        filter1 = self.collector.gen_filter(value1=2)
        actual = self.collector.extend_filter(filter1, lop='or')
        self.assertEqual(filter1, actual)

    def test_try_extend_empty_filter(self):
        actual = self.collector.extend_filter()
        self.assertEqual({}, actual)
        actual = self.collector.extend_filter(actual, actual)
        self.assertEqual({}, actual)

    def test_try_extend_filter_with_none(self):
        filter1 = self.collector.gen_filter(value1=2)
        actual = self.collector.extend_filter(filter1, None)
        self.assertEqual(filter1, actual)

    def test_generate_two_logical_ops(self):
        filter1 = self.collector.gen_filter(value1=2, value2=3)
        filter2 = self.collector.gen_filter(cop='<=', value3=1)
        actual = self.collector.extend_filter(filter1, filter2, lop='or')
        expected = {'or': [{
            'and': [{
                '=': {
                    'value1': 2
                }}, {
                '=': {
                    'value2': 3
                }}]}, {
            '<=': {
                'value3': 1
            }}]}
        self.assertEqual(expected, actual)

    def test_gen_filter_parameters(self):
        actual = self.collector.gen_filter(
            cop='>',
            lop='or',
            value1=2,
            value2=3)
        expected = {'or': [{
            '>': {
                'value1': 2
            }}, {
            '>': {
                'value2': 3
            }}]}
        self.assertEqual(expected, actual)

    def test_extend_filter_parameters(self):
        actual = self.collector.extend_filter(
            ['dummy1'],
            ['dummy2'],
            lop='or')
        expected = {'or': ['dummy1', 'dummy2']}
        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from cloudkitty import transformer


class GnocchiTransformer(transformer.BaseTransformer):
    compute_map = {
        'instance_id': ['id'],
        'name': ['display_name'],
        'flavor_id': ['flavor_id'],
        'image_id': lambda x, y: x.get_image_id(y),
    }
    image_map = {
        'container_format': ['container_format'],
        'disk_format': ['disk_format'],
    }
    volume_map = {
        'name': ['display_name'],
        'volume_type': ['volume_type'],
    }
    network_map = {
        'name': ['name'],
    }

    def _generic_strip(self, data):
        res_data = {
            'resource_id': data['id'],
            'project_id': data['project_id'],
            'user_id': data['user_id'],
            'metrics': data['metrics']}
        return res_data

    @staticmethod
    def get_image_id(data):
        image_ref = data.get('image_ref', None)
        return image_ref.rpartition('/')[-1] if image_ref else None

    def strip_resource_data(self, res_type, res_data):
        result = self._generic_strip(res_data)
        stripped_data = super(GnocchiTransformer, self).strip_resource_data(
            res_type,
            res_data)
        result.update(stripped_data)
        return result

    def get_metadata(self, res_type):
        """Return list of metadata available after transformation for

        given resource type.
        """

        class FakeData(dict):
            """FakeData object."""

            def __getitem__(self, item):
                try:
                    return super(FakeData, self).__getitem__(item)
                except KeyError:
                    return item

            def get(self, item, default=None):
                return super(FakeData, self).get(item, item)

        # list of metadata is built by applying the generic strip_resource_data
        # function to a fake data object

        fkdt = FakeData()
        res_data = self.strip_resource_data(res_type, fkdt)
        return res_data.keys()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

from oslo_config import cfg
from oslo_db import api as db_api
import six

_BACKEND_MAPPING = {'sqlalchemy': 'cloudkitty.rating.hash.db.sqlalchemy.api'}
IMPL = db_api.DBAPI.from_config(cfg.CONF,
                                backend_mapping=_BACKEND_MAPPING,
                                lazy=True)


def get_instance():
    """Return a DB API instance."""
    return IMPL


class BaseHashMapError(Exception):
    """Base class for HashMap errors."""


class ClientHashMapError(BaseHashMapError):
    """Base class for client side errors."""


class NoSuchService(ClientHashMapError):
    """Raised when the service doesn't exist."""

    def __init__(self, name=None, uuid=None):
        super(NoSuchService, self).__init__(
            "No such service: %s (UUID: %s)" % (name, uuid))
        self.name = name
        self.uuid = uuid


class NoSuchField(ClientHashMapError):
    """Raised when the field doesn't exist for the service."""

    def __init__(self, uuid):
        super(NoSuchField, self).__init__(
            "No such field: %s" % uuid)
        self.uuid = uuid


class NoSuchGroup(ClientHashMapError):
    """Raised when the group doesn't exist."""

    def __init__(self, name=None, uuid=None):
        super(NoSuchGroup, self).__init__(
            "No such group: %s (UUID: %s)" % (name, uuid))
        self.name = name
        self.uuid = uuid


class NoSuchMapping(ClientHashMapError):
    """Raised when the mapping doesn't exist."""

    def __init__(self, uuid):
        msg = ("No such mapping: %s" % uuid)
        super(NoSuchMapping, self).__init__(msg)
        self.uuid = uuid


class NoSuchThreshold(ClientHashMapError):
    """Raised when the threshold doesn't exist."""

    def __init__(self, uuid):
        msg = ("No such threshold: %s" % uuid)
        super(NoSuchThreshold, self).__init__(msg)
        self.uuid = uuid


class NoSuchType(ClientHashMapError):
    """Raised when a mapping type is not handled."""

    def __init__(self, map_type):
        msg = ("No mapping type: %s"
               % (map_type))
        super(NoSuchType, self).__init__(msg)
        self.map_type = map_type


class ServiceAlreadyExists(ClientHashMapError):
    """Raised when the service already exists."""

    def __init__(self, name, uuid):
        super(ServiceAlreadyExists, self).__init__(
            "Service %s already exists (UUID: %s)" % (name, uuid))
        self.name = name
        self.uuid = uuid


class FieldAlreadyExists(ClientHashMapError):
    """Raised when the field already exists."""

    def __init__(self, field, uuid):
        super(FieldAlreadyExists, self).__init__(
            "Field %s already exists (UUID: %s)" % (field, uuid))
        self.field = field
        self.uuid = uuid


class GroupAlreadyExists(ClientHashMapError):
    """Raised when the group already exists."""

    def __init__(self, name, uuid):
        super(GroupAlreadyExists, self).__init__(
            "Group %s already exists (UUID: %s)" % (name, uuid))
        self.name = name
        self.uuid = uuid


class MappingAlreadyExists(ClientHashMapError):
    """Raised when the mapping already exists."""

    def __init__(self,
                 mapping,
                 parent_id=None,
                 parent_type=None,
                 uuid=None,
                 tenant_id=None):
        # TODO(sheeprine): UUID is deprecated
        parent_id = parent_id if parent_id else uuid
        super(MappingAlreadyExists, self).__init__(
            "Mapping '%s' already exists for %s '%s', tenant: '%s'"
            % (mapping, parent_type, parent_id, tenant_id))
        self.mapping = mapping
        self.uuid = parent_id
        self.parent_id = parent_id
        self.parent_type = parent_type
        self.tenant_id = tenant_id


class ThresholdAlreadyExists(ClientHashMapError):
    """Raised when the threshold already exists."""

    def __init__(self,
                 threshold,
                 parent_id=None,
                 parent_type=None,
                 uuid=None,
                 tenant_id=None):
        # TODO(sheeprine): UUID is deprecated
        parent_id = parent_id if parent_id else uuid
        super(ThresholdAlreadyExists, self).__init__(
            "Threshold '%s' already exists for %s '%s', tenant: '%s'"
            % (threshold, parent_type, parent_id, tenant_id))
        self.threshold = threshold
        self.uuid = parent_id
        self.parent_id = parent_id
        self.parent_type = parent_type
        self.tenant_id = tenant_id


class MappingHasNoGroup(ClientHashMapError):
    """Raised when the mapping is not attached to a group."""

    def __init__(self, uuid):
        super(MappingHasNoGroup, self).__init__(
            "Mapping has no group (UUID: %s)" % uuid)
        self.uuid = uuid


class ThresholdHasNoGroup(ClientHashMapError):
    """Raised when the threshold is not attached to a group."""

    def __init__(self, uuid):
        super(ThresholdHasNoGroup, self).__init__(
            "Threshold has no group (UUID: %s)" % uuid)
        self.uuid = uuid


@six.add_metaclass(abc.ABCMeta)
class HashMap(object):
    """Base class for hashmap configuration."""

    @abc.abstractmethod
    def get_migration(self):
        """Return a migrate manager.

        """

    @abc.abstractmethod
    def get_service(self, name=None, uuid=None):
        """Return a service object.

        :param name: Filter on a service name.
        :param uuid: The uuid of the service to get.
        """

    @abc.abstractmethod
    def get_field(self, uuid=None, service_uuid=None, name=None):
        """Return a field object.

        :param uuid: UUID of the field to get.
        :param service_uuid: UUID of the service to filter on. (Used with name)
        :param name: Name of the field to filter on. (Used with service_uuid)
        """

    @abc.abstractmethod
    def get_group(self, uuid=None, name=None):
        """Return a group object.

        :param uuid: UUID of the group to get.
        :param name: Name of the group to get.
        """

    @abc.abstractmethod
    def get_mapping(self, uuid):
        """Return a mapping object.

        :param uuid: UUID of the mapping to get.
        """

    @abc.abstractmethod
    def get_threshold(self, uuid):
        """Return a threshold object.

        :param uuid: UUID of the threshold to get.
        """

    @abc.abstractmethod
    def list_services(self):
        """Return an UUID list of every service.

        """

    @abc.abstractmethod
    def list_fields(self, service_uuid):
        """Return an UUID list of every field in a service.

        :param service_uuid: The service UUID to filter on.
        """

    @abc.abstractmethod
    def list_groups(self):
        """Return an UUID list of every group.

        """

    @abc.abstractmethod
    def list_mappings(self,
                      service_uuid=None,
                      field_uuid=None,
                      group_uuid=None,
                      no_group=False,
                      **kwargs):
        """Return an UUID list of every mapping.

        :param service_uuid: The service to filter on.
        :param field_uuid: The field to filter on.
        :param group_uuid: The group to filter on.
        :param no_group: Filter on mappings without a group.
        :param tenant_uuid: The tenant to filter on.

        :return list(str): List of mappings' UUID.
        """

    @abc.abstractmethod
    def list_thresholds(self,
                        service_uuid=None,
                        field_uuid=None,
                        group_uuid=None,
                        no_group=False,
                        **kwargs):
        """Return an UUID list of every threshold.

        :param service_uuid: The service to filter on.
        :param field_uuid: The field to filter on.
        :param group_uuid: The group to filter on.
        :param no_group: Filter on thresholds without a group.
        :param tenant_uuid: The tenant to filter on.

        :return list(str): List of thresholds' UUID.
        """

    @abc.abstractmethod
    def create_service(self, name):
        """Create a new service.

        :param name: Name of the service to create.
        """

    @abc.abstractmethod
    def create_field(self, service_uuid, name):
        """Create a new field.

        :param service_uuid: UUID of the parent service.
        :param name: Name of the field.
        """

    @abc.abstractmethod
    def create_group(self, name):
        """Create a new group.

        :param name: The name of the group.
        """

    @abc.abstractmethod
    def create_mapping(self,
                       cost,
                       map_type='rate',
                       value=None,
                       service_id=None,
                       field_id=None,
                       group_id=None,
                       tenant_id=None):
        """Create a new service/field mapping.

        :param cost: Rating value to apply to this mapping.
        :param map_type: The type of rating rule.
        :param value: Value of the field this mapping is applying to.
        :param service_id: Service the mapping is applying to.
        :param field_id: Field the mapping is applying to.
        :param group_id: The group of calculations to apply.
        :param tenant_id: The tenant to apply calculations to.
        """

    @abc.abstractmethod
    def create_threshold(self,
                         cost,
                         map_type='rate',
                         level=None,
                         service_id=None,
                         field_id=None,
                         group_id=None,
                         tenant_id=None):
        """Create a new service/field threshold.

        :param cost: Rating value to apply to this threshold.
        :param map_type: The type of rating rule.
        :param level: Level of the field this threshold is applying to.
        :param service_id: Service the threshold is applying to.
        :param field_id: Field the threshold is applying to.
        :param group_id: The group of calculations to apply.
        :param tenant_id: The tenant to apply calculations to.
        """

    @abc.abstractmethod
    def update_mapping(self, uuid, **kwargs):
        """Update a mapping.

        :param uuid UUID of the mapping to modify.
        :param cost: Rating value to apply to this mapping.
        :param map_type: The type of rating rule.
        :param value: Value of the field this mapping is applying to.
        :param group_id: The group of calculations to apply.
        """

    @abc.abstractmethod
    def update_threshold(self, uuid, **kwargs):
        """Update a mapping.

        :param uuid UUID of the threshold to modify.
        :param cost: Rating value to apply to this threshold.
        :param map_type: The type of rating rule.
        :param level: Level of the field this threshold is applying to.
        :param group_id: The group of calculations to apply.
        """

    @abc.abstractmethod
    def delete_service(self, name=None, uuid=None):
        """Delete a service recursively.

        :param name: Name of the service to delete.
        :param uuid: UUID of the service to delete.
        """

    @abc.abstractmethod
    def delete_field(self, uuid):
        """Delete a field recursively.

        :param uuid UUID of the field to delete.
        """

    def delete_group(self, uuid, recurse=True):
        """Delete a group and all mappings recursively.

        :param uuid: UUID of the group to delete.
        :param recurse: Delete attached mappings recursively.
        """

    @abc.abstractmethod
    def delete_mapping(self, uuid):
        """Delete a mapping

        :param uuid: UUID of the mapping to delete.
        """
    @abc.abstractmethod
    def delete_threshold(self, uuid):
        """Delete a threshold

        :param uuid: UUID of the threshold to delete.
        """



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
"""
Time calculations functions

We're mostly using oslo_utils for time calculations but we're encapsulating it
to ease maintenance in case of library modifications.
"""
import calendar
import contextlib
import datetime
import decimal
import fractions
import shutil
import six
import sys
import tempfile
import yaml

from oslo_config import cfg
from oslo_log import log as logging

from oslo_utils import timeutils
from six import moves
from stevedore import extension


_ISO8601_TIME_FORMAT_SUBSECOND = '%Y-%m-%dT%H:%M:%S.%f'
_ISO8601_TIME_FORMAT = '%Y-%m-%dT%H:%M:%S'

LOG = logging.getLogger(__name__)

collect_opts = [
    cfg.StrOpt('collector',
               default='gnocchi',
               deprecated_for_removal=True,
               help='Data collector.'),
    cfg.IntOpt('window',
               default=1800,
               deprecated_for_removal=True,
               help='Number of samples to collect per call.'),
    cfg.IntOpt('period',
               default=3600,
               deprecated_for_removal=True,
               help='Rating period in seconds.'),
    cfg.IntOpt('wait_periods',
               default=2,
               deprecated_for_removal=True,
               help='Wait for N periods before collecting new data.'),
    cfg.ListOpt('services',
                default=[
                    'compute',
                    'volume',
                    'network.bw.in',
                    'network.bw.out',
                    'network.floating',
                    'image',
                ],
                deprecated_for_removal=True,
                help='Services to monitor.'),
    cfg.StrOpt('metrics_conf',
               default='/etc/cloudkitty/metrics.yml',
               help='Metrology configuration file.'),
]
CONF = cfg.CONF
CONF.register_opts(collect_opts, 'collect')


def isotime(at=None, subsecond=False):
    """Stringify time in ISO 8601 format."""

    # Python provides a similar instance method for datetime.datetime objects
    # called isoformat(). The format of the strings generated by isoformat()
    # have a couple of problems:
    # 1) The strings generated by isotime are used in tokens and other public
    #    APIs that we can't change without a deprecation period. The strings
    #    generated by isoformat are not the same format, so we can't just
    #    change to it.
    # 2) The strings generated by isoformat do not include the microseconds if
    #    the value happens to be 0. This will likely show up as random failures
    #    as parsers may be written to always expect microseconds, and it will
    #    parse correctly most of the time.

    if not at:
        at = timeutils.utcnow()
    st = at.strftime(_ISO8601_TIME_FORMAT
                     if not subsecond
                     else _ISO8601_TIME_FORMAT_SUBSECOND)
    tz = at.tzinfo.tzname(None) if at.tzinfo else 'UTC'
    st += ('Z' if tz == 'UTC' else tz)
    return st


def iso8601_from_timestamp(timestamp, microsecond=False):
    """Returns an iso8601 formatted date from timestamp"""

    # Python provides a similar instance method for datetime.datetime
    # objects called isoformat() and utcfromtimestamp(). The format
    # of the strings generated by isoformat() and utcfromtimestamp()
    # have a couple of problems:
    # 1) The method iso8601_from_timestamp in oslo_utils is realized
    #    by isotime, the strings generated by isotime are used in
    #    tokens and other public APIs that we can't change without a
    #    deprecation period. The strings generated by isoformat are
    #    not the same format, so we can't just change to it.
    # 2) The strings generated by isoformat() and utcfromtimestamp()
    #    do not include the microseconds if the value happens to be 0.
    #    This will likely show up as random failures as parsers may be
    #    written to always expect microseconds, and it will parse
    #    correctly most of the time.

    return isotime(datetime.datetime.utcfromtimestamp(timestamp), microsecond)


def dt2ts(orig_dt):
    """Translate a datetime into a timestamp."""
    return calendar.timegm(orig_dt.timetuple())


def iso2dt(iso_date):
    """iso8601 format to datetime."""
    iso_dt = timeutils.parse_isotime(iso_date)
    trans_dt = timeutils.normalize_time(iso_dt)
    return trans_dt


def ts2dt(timestamp):
    """timestamp to datetime format."""
    if not isinstance(timestamp, float):
        timestamp = float(timestamp)
    return datetime.datetime.utcfromtimestamp(timestamp)


def ts2iso(timestamp):
    """timestamp to is8601 format."""
    if not isinstance(timestamp, float):
        timestamp = float(timestamp)
    return iso8601_from_timestamp(timestamp)


def dt2iso(orig_dt):
    """datetime to is8601 format."""
    return isotime(orig_dt)


def utcnow():
    """Returns a datetime for the current utc time."""
    return timeutils.utcnow()


def utcnow_ts():
    """Returns a timestamp for the current utc time."""
    return timeutils.utcnow_ts()


def get_month_days(dt):
    return calendar.monthrange(dt.year, dt.month)[1]


def add_days(base_dt, days, stay_on_month=True):
    if stay_on_month:
        max_days = get_month_days(base_dt)
        if days > max_days:
            return get_month_end(base_dt)
    return base_dt + datetime.timedelta(days=days)


def add_month(dt, stay_on_month=True):
    next_month = get_next_month(dt)
    return add_days(next_month, dt.day, stay_on_month)


def sub_month(dt, stay_on_month=True):
    prev_month = get_last_month(dt)
    return add_days(prev_month, dt.day, stay_on_month)


def get_month_start(dt=None):
    if not dt:
        dt = utcnow()
    month_start = datetime.datetime(dt.year, dt.month, 1)
    return month_start


def get_month_start_timestamp(dt=None):
    return dt2ts(get_month_start(dt))


def get_month_end(dt=None):
    month_start = get_month_start(dt)
    days_of_month = get_month_days(month_start)
    month_end = month_start.replace(day=days_of_month)
    return month_end


def get_last_month(dt=None):
    if not dt:
        dt = utcnow()
    month_end = get_month_start(dt) - datetime.timedelta(days=1)
    return get_month_start(month_end)


def get_next_month(dt=None):
    month_end = get_month_end(dt)
    next_month = month_end + datetime.timedelta(days=1)
    return next_month


def get_next_month_timestamp(dt=None):
    return dt2ts(get_next_month(dt))


def refresh_stevedore(namespace=None):
    """Trigger reload of entry points.

    Useful to have dynamic loading/unloading of stevedore modules.
    """
    # NOTE(sheeprine): pkg_resources doesn't support reload on python3 due to
    # defining basestring which is still there on reload hence executing
    # python2 related code.
    try:
        del sys.modules['pkg_resources'].basestring
    except AttributeError:
        # python2, do nothing
        pass
    # Force working_set reload
    moves.reload_module(sys.modules['pkg_resources'])
    # Clear stevedore cache
    cache = extension.ExtensionManager.ENTRY_POINT_CACHE
    if namespace:
        if namespace in cache:
            del cache[namespace]
    else:
        cache.clear()


def check_time_state(timestamp=None, period=0, wait_time=0):
    if not timestamp:
        return get_month_start_timestamp()

    now = utcnow_ts()
    next_timestamp = timestamp + period
    if next_timestamp + wait_time < now:
        return next_timestamp
    return 0


def get_metrics_conf(conf_path):
    """Return loaded yaml metrology configuration.

    In case of empty /etc/cloudkitty folder,
    a fallback is done on the former deprecated oslo config method.
    """
    res = None
    try:
        with open(conf_path) as conf:
            res = yaml.safe_load(conf)
            res = res[0]
    except Exception as exc:
        LOG.warning('Error when trying to retrieve yaml metrology conf file.')
        LOG.warning(exc)
        LOG.warning('Fallback on the deprecated oslo config method.')

        try:
            res = {key: val for key, val in CONF.collect.items()}
        except Exception as exc:
            err_msg = 'Error when trying to retrieve ' \
                      'deprecated oslo config method.'
            LOG.error(err_msg)
            LOG.error(exc)

    return res


@contextlib.contextmanager
def tempdir(**kwargs):
    tmpdir = tempfile.mkdtemp(**kwargs)
    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            LOG.debug('Could not remove tmpdir: %s',
                      six.text_type(e))


def num2decimal(num):
    """Converts a number into a decimal.Decimal.

    The number may be an str in float, int or fraction format;
    a fraction.Fraction, a decimal.Decimal, an int or a float.
    """
    if isinstance(num, decimal.Decimal):
        return num
    if isinstance(num, str):
        if '/' in num:
            num = float(fractions.Fraction(num))
    if isinstance(num, fractions.Fraction):
        num = float(num)
    return decimal.Decimal(num)


def convert_unit(value, factor=1, offset=0):
    """Return converted value depending on the provided factor and offset."""
    return num2decimal(value) * num2decimal(factor) + num2decimal(offset)


def flat_dict(item, parent=None):
    """Returns a flat version of the nested dict item"""
    if not parent:
        parent = dict()
    for k, val in item.items():
        if isinstance(val, dict):
            parent = flat_dict(val, parent)
        else:
            parent[k] = val
    return parent



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal
import json
import os

from cloudkitty import writer


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        return super(DecimalEncoder, self).default(o)


class OSRFBackend(writer.BaseReportWriter):
    """OpenStack Report Format Writer:

        Generates report in native format (json)
    """
    report_type = 'osrf'

    def _gen_filename(self, timeframe):
        filename = '{}-osrf-{}-{:02d}.json'.format(self._tenant_id,
                                                   timeframe.year,
                                                   timeframe.month)
        if self._basepath:
            filename = os.path.join(self._basepath, filename)
        return filename

    def _open(self):
        filename = self._gen_filename(self.usage_start_dt)
        self._report = self._backend(filename, 'rb+')
        self._report.seek(0, 2)
        if self._report.tell():
            self._recover_state()
        else:
            self._report.seek(0)

    def _write_header(self):
        self._report.write('[')
        self._report.flush()

    def _write_total(self):
        total = {'total': self.total}
        self._report.write(json.dumps(total, cls=DecimalEncoder))
        self._report.write(']')
        self._report.flush()

    def _recover_state(self):
        # Search for last comma
        self._report.seek(0, 2)
        max_idx = self._report.tell()
        if max_idx > 2000:
            max_idx = 2000
        hay = ''
        for idx in range(10, max_idx, 10):
            self._report.seek(-idx, 2)
            hay = self._report.read()
            if hay.count(','):
                break
        last_comma = hay.rfind(',')
        if last_comma > -1:
            last_comma -= len(hay)
        else:
            raise RuntimeError('Unable to recover file state.')
        self._report.seek(last_comma, 2)
        self._report.write(', ')
        self._report.truncate()

    def _close_file(self):
        if self._report is not None:
            self._recover_state()
            self._write_total()
            self._report.close()

    def _write(self):
        data = {}
        data['period'] = {'begin': self.usage_start_dt.isoformat(),
                          'end': self.usage_end_dt.isoformat()}
        data['usage'] = self._usage_data

        self._report.write(json.dumps(data, cls=DecimalEncoder))
        self._report.write(', ')
        self._report.flush()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from pecan import rest

from cloudkitty.api.v1.controllers import collector as collector_api
from cloudkitty.api.v1.controllers import info as info_api
from cloudkitty.api.v1.controllers import rating as rating_api
from cloudkitty.api.v1.controllers import report as report_api
from cloudkitty.api.v1.controllers import storage as storage_api


class V1Controller(rest.RestController):
    """API version 1 controller.

    """

    billing = rating_api.RatingController()
    collector = collector_api.CollectorController()
    rating = rating_api.RatingController()
    report = report_api.ReportController()
    storage = storage_api.StorageController()
    info = info_api.InfoController()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from cloudkitty import transformer


class CloudKittyFormatTransformer(transformer.BaseTransformer):
    def format_item(self, desc, unit, qty=1.0):
        data = {}
        data['desc'] = desc
        data['vol'] = {'unit': unit, 'qty': qty}

        return data

    def format_service(self, service, items):
        data = {}
        data[service] = items

        return data



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial migration

Revision ID: 3dd7e13527f3
Revises: None
Create Date: 2015-03-10 13:06:41.067563

"""

# revision identifiers, used by Alembic.
revision = '3dd7e13527f3'
down_revision = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'hashmap_services',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('service_id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name'),
        sa.UniqueConstraint('service_id'),
        mysql_charset='utf8',
        mysql_engine='InnoDB')
    op.create_table(
        'hashmap_fields',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('field_id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('service_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['service_id'],
            ['hashmap_services.id'],
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('field_id'),
        sa.UniqueConstraint('field_id', 'name', name='uniq_field'),
        sa.UniqueConstraint(
            'service_id',
            'name',
            name='uniq_map_service_field'),
        mysql_charset='utf8',
        mysql_engine='InnoDB')
    op.create_table(
        'hashmap_groups',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('group_id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('group_id'),
        sa.UniqueConstraint('name'),
        mysql_charset='utf8',
        mysql_engine='InnoDB')
    op.create_table(
        'hashmap_maps',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('mapping_id', sa.String(length=36), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=True),
        sa.Column('cost', sa.Numeric(20, 8), nullable=False),
        sa.Column(
            'map_type',
            sa.Enum('flat', 'rate', name='enum_map_type'),
            nullable=False),
        sa.Column('service_id', sa.Integer(), nullable=True),
        sa.Column('field_id', sa.Integer(), nullable=True),
        sa.Column('group_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['field_id'],
            ['hashmap_fields.id'],
            ondelete='CASCADE'),
        sa.ForeignKeyConstraint(
            ['group_id'],
            ['hashmap_groups.id'],
            ondelete='SET NULL'),
        sa.ForeignKeyConstraint(
            ['service_id'],
            ['hashmap_services.id'],
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('mapping_id'),
        sa.UniqueConstraint(
            'value',
            'field_id',
            name='uniq_field_mapping'),
        sa.UniqueConstraint(
            'value',
            'service_id',
            name='uniq_service_mapping'),
        mysql_charset='utf8',
        mysql_engine='InnoDB')



# -*- coding: utf-8 -*-
# Copyright 2016 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import copy

from cloudkitty import tests
from cloudkitty.tests import samples
from cloudkitty.tests import transformers as t_transformers

TRANS_METADATA = {
    'availability_zone': 'nova',
    'flavor': 'm1.nano',
    'image_id': 'f5600101-8fa2-4864-899e-ebcb7ed6b568',
    'memory': '64',
    'name': 'prod1',
    'vcpus': '1'}


class TransformerBaseTest(tests.TestCase):
    def test_strip_resource_on_dict(self):
        metadata = copy.deepcopy(samples.COMPUTE_METADATA)
        t_test = t_transformers.Transformer()
        result = t_test.strip_resource_data('compute', metadata)
        self.assertEqual(TRANS_METADATA, result)

    def test_strip_resource_with_no_rules(self):
        metadata = copy.deepcopy(samples.COMPUTE_METADATA)
        t_test = t_transformers.Transformer()
        result = t_test.strip_resource_data('unknown', metadata)
        self.assertEqual(samples.COMPUTE_METADATA, result)

    def test_strip_resource_with_func(self):
        metadata = {'test': 'dummy'}
        t_test = t_transformers.Transformer()
        result = t_test.strip_resource_data('test', metadata)
        self.assertEqual({'test': 'ok'}, result)

    def test_strip_resource_with_stripping_function(self):
        metadata = {}
        t_test = t_transformers.Transformer()
        result = t_test.strip_resource_data('network', metadata)
        self.assertEqual({'test': 'ok'}, result)

    def test_strip_resource_with_subitem(self):
        test_obj = t_transformers.EmptyClass()
        test_obj.metadata = copy.deepcopy(samples.COMPUTE_METADATA)
        t_test = t_transformers.TransformerMeta()
        result = t_test.strip_resource_data('compute', test_obj)
        self.assertEqual(TRANS_METADATA, result)

    def test_strip_resource_with_attributes(self):
        test_obj = t_transformers.EmptyClass()
        test_obj.metadata = t_transformers.ClassWithAttr()
        t_test = t_transformers.TransformerMeta()
        result = t_test.strip_resource_data('compute', test_obj)
        self.assertEqual(TRANS_METADATA, result)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import datetime
import decimal

import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1.datamodels import report as report_models
from cloudkitty.common import policy
from cloudkitty import utils as ck_utils


class ReportController(rest.RestController):
    """REST Controller managing the reporting.

    """

    _custom_actions = {
        'total': ['GET'],
        'tenants': ['GET'],
        'summary': ['GET']
    }

    @wsme_pecan.wsexpose([wtypes.text],
                         datetime.datetime,
                         datetime.datetime)
    def tenants(self, begin=None, end=None):
        """Return the list of rated tenants.

        """
        policy.authorize(pecan.request.context, 'report:list_tenants', {})

        if not begin:
            begin = ck_utils.get_month_start()
        if not end:
            end = ck_utils.get_next_month()

        storage = pecan.request.storage_backend
        tenants = storage.get_tenants(begin, end)
        return tenants

    @wsme_pecan.wsexpose(decimal.Decimal,
                         datetime.datetime,
                         datetime.datetime,
                         wtypes.text,
                         wtypes.text,
                         bool)
    def total(self, begin=None, end=None, tenant_id=None, service=None,
              all_tenants=False):
        """Return the amount to pay for a given period.

        """
        if not begin:
            begin = ck_utils.get_month_start()
        if not end:
            end = ck_utils.get_next_month()

        if all_tenants:
            tenant_id = None
        else:
            tenant_context = pecan.request.context.tenant
            tenant_id = tenant_context if not tenant_id else tenant_id
        policy.authorize(pecan.request.context, 'report:get_total',
                         {"tenant_id": tenant_id})

        storage = pecan.request.storage_backend
        # FIXME(sheeprine): We should filter on user id.
        # Use keystone token information by default but make it overridable and
        # enforce it by policy engine
        total = storage.get_total(begin, end, tenant_id, service)

        # TODO(Aaron): `get_total` return a list of dict,
        # Get value of rate from index[0]
        total = total[0].get('rate', decimal.Decimal('0'))
        return total if total else decimal.Decimal('0')

    @wsme_pecan.wsexpose(report_models.SummaryCollectionModel,
                         datetime.datetime,
                         datetime.datetime,
                         wtypes.text,
                         wtypes.text,
                         wtypes.text,
                         bool)
    def summary(self, begin=None, end=None, tenant_id=None,
                service=None, groupby=None, all_tenants=False):
        """Return the summary to pay for a given period.

        """
        if not begin:
            begin = ck_utils.get_month_start()
        if not end:
            end = ck_utils.get_next_month()

        if all_tenants:
            tenant_id = None
        else:
            tenant_context = pecan.request.context.tenant
            tenant_id = tenant_context if not tenant_id else tenant_id
        policy.authorize(pecan.request.context, 'report:get_summary',
                         {"tenant_id": tenant_id})
        storage = pecan.request.storage_backend

        summarymodels = []
        results = storage.get_total(begin, end, tenant_id, service,
                                    groupby=groupby)
        for result in results:
            summarymodel = report_models.SummaryModel(**result)
            summarymodels.append(summarymodel)

        return report_models.SummaryCollectionModel(summary=summarymodels)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from __future__ import print_function

from oslo_config import cfg
from oslo_utils import importutils as i_utils

from cloudkitty import config  # noqa
from cloudkitty import service
from cloudkitty import storage
from cloudkitty import utils as ck_utils
from cloudkitty import write_orchestrator

CONF = cfg.CONF
CONF.import_opt('period', 'cloudkitty.collector', 'collect')
CONF.import_opt('backend', 'cloudkitty.config', 'output')
CONF.import_opt('basepath', 'cloudkitty.config', 'output')
STORAGES_NAMESPACE = 'cloudkitty.storage.backends'


class DBCommand(object):

    def __init__(self):
        self._storage = None
        self._output = None
        self._load_storage_backend()
        self._load_output_backend()

    def _load_storage_backend(self):
        self._storage = storage.get_storage()

    def _load_output_backend(self):
        backend = i_utils.import_class(CONF.output.backend)
        self._output = backend

    def generate(self):
        if not CONF.command.tenant:
            if not CONF.command.begin:
                CONF.command.begin = ck_utils.get_month_start()
            if not CONF.command.end:
                CONF.command.end = ck_utils.get_next_month()
            tenants = self._storage.get_tenants(CONF.command.begin,
                                                CONF.command.end)
        else:
            tenants = [CONF.command.tenant]
        for tenant in tenants:
            wo = write_orchestrator.WriteOrchestrator(self._output,
                                                      tenant,
                                                      self._storage,
                                                      CONF.output.basepath)
            wo.init_writing_pipeline()
            if not CONF.command.begin:
                wo.restart_month()
            wo.process()

    def tenants_list(self):
        if not CONF.command.begin:
            CONF.command.begin = ck_utils.get_month_start()
        if not CONF.command.end:
            CONF.command.end = ck_utils.get_next_month()
        tenants = self._storage.get_tenants(CONF.command.begin,
                                            CONF.command.end)
        print('Tenant list:')
        for tenant in tenants:
            print(tenant)


def call_generate(command_object):
    command_object.generate()


def call_tenants_list(command_object):
    command_object.tenants_list()


def add_command_parsers(subparsers):
    parser = subparsers.add_parser('generate')
    parser.set_defaults(func=call_generate)
    parser.add_argument('--tenant', nargs='?')
    parser.add_argument('--begin', nargs='?')
    parser.add_argument('--end', nargs='?')

    parser = subparsers.add_parser('tenants_list')
    parser.set_defaults(func=call_tenants_list)
    parser.add_argument('--begin', nargs='?')
    parser.add_argument('--end', nargs='?')


command_opt = cfg.SubCommandOpt('command',
                                title='Command',
                                help='Available commands',
                                handler=add_command_parsers)

CONF.register_cli_opt(command_opt)


def main():
    service.prepare_service()
    command_object = DBCommand()
    CONF.command.func(command_object)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from wsme import types as wtypes

from cloudkitty.api.v1 import types as ck_types


class Service(wtypes.Base):
    """Type describing a service.

    A service is directly mapped to the usage key, the collected service.
    """

    service_id = wtypes.wsattr(ck_types.UuidType(), mandatory=False)
    """UUID of the service."""

    name = wtypes.wsattr(wtypes.text, mandatory=True)
    """Name of the service."""

    @classmethod
    def sample(cls):
        sample = cls(service_id='a733d0e1-1ec9-4800-8df8-671e4affd017',
                     name='compute')
        return sample


class ServiceCollection(wtypes.Base):
    """Type describing a list of services."""

    services = [Service]
    """List of services."""

    @classmethod
    def sample(cls):
        sample = Service.sample()
        return cls(services=[sample])



# -*- coding: utf-8 -*-
# Copyright 2017 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Luka Peschke
#
from oslo_db.sqlalchemy import models
import sqlalchemy
from sqlalchemy.ext import declarative

Base = declarative.declarative_base()


class TenantState(Base, models.ModelBase):
    """A tenant state.

    """
    __table_args__ = {'mysql_charset': "utf8",
                      'mysql_engine': "InnoDB"}
    __tablename__ = 'hybrid_storage_states'

    id = sqlalchemy.Column(sqlalchemy.Integer,
                           primary_key=True)
    tenant_id = sqlalchemy.Column(sqlalchemy.String(32),
                                  nullable=False)
    state = sqlalchemy.Column(sqlalchemy.DateTime,
                              nullable=False)



# -*- coding: utf-8 -*-
# !/usr/bin/env python
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from keystoneauth1 import loading as ks_loading
from keystoneclient import client as kclient
from keystoneclient import discover
from keystoneclient import exceptions
from oslo_config import cfg

from cloudkitty import tenant_fetcher

KEYSTONE_FETCHER_OPTS = 'keystone_fetcher'
keystone_common_opts = ks_loading.get_auth_common_conf_options()
keystone_fetcher_opts = [
    cfg.StrOpt('keystone_version',
               default='2',
               help='Keystone version to use.'), ]

cfg.CONF.register_opts(keystone_common_opts, KEYSTONE_FETCHER_OPTS)
cfg.CONF.register_opts(keystone_fetcher_opts, KEYSTONE_FETCHER_OPTS)
ks_loading.register_session_conf_options(
    cfg.CONF,
    KEYSTONE_FETCHER_OPTS)
ks_loading.register_auth_conf_options(
    cfg.CONF,
    KEYSTONE_FETCHER_OPTS)
CONF = cfg.CONF


class KeystoneFetcher(tenant_fetcher.BaseFetcher):
    """Keystone tenants fetcher."""

    def __init__(self):
        self.auth = ks_loading.load_auth_from_conf_options(
            CONF,
            KEYSTONE_FETCHER_OPTS)
        self.session = ks_loading.load_session_from_conf_options(
            CONF,
            KEYSTONE_FETCHER_OPTS,
            auth=self.auth)
        self.admin_ks = kclient.Client(
            version=CONF.keystone_fetcher.keystone_version,
            session=self.session,
            auth_url=self.auth.auth_url)

    def get_tenants(self):
        keystone_version = discover.normalize_version_number(
            CONF.keystone_fetcher.keystone_version)
        auth_dispatch = {(3,): ('project', 'projects', 'list'),
                         (2,): ('tenant', 'tenants', 'roles_for_user')}
        for auth_version, auth_version_mapping in auth_dispatch.items():
            if discover.version_match(auth_version, keystone_version):
                return self._do_get_tenants(auth_version_mapping)
        msg = "Keystone version you've specified is not supported"
        raise exceptions.VersionNotAvailable(msg)

    def _do_get_tenants(self, auth_version_mapping):
        tenant_attr, tenants_attr, role_func = auth_version_mapping
        tenant_list = getattr(self.admin_ks, tenants_attr).list()
        my_user_id = self.session.get_user_id()
        for tenant in tenant_list[:]:
            roles = getattr(self.admin_ks.roles, role_func)(
                **{'user': my_user_id,
                   tenant_attr: tenant})
            if 'rating' not in [role.name for role in roles]:
                tenant_list.remove(tenant)
        return [tenant.id for tenant in tenant_list]



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import mock

from cloudkitty.db import api as ck_db_api
from cloudkitty import tests


class FakeRPCClient(object):
    def __init__(self, namespace=None, fanout=False):
        self._queue = []
        self._namespace = namespace
        self._fanout = fanout

    def prepare(self, namespace=None, fanout=False):
        self._namespace = namespace
        self._fanout = fanout
        return self

    def cast(self, ctx, data, **kwargs):
        cast_data = {'ctx': ctx,
                     'data': data}
        cast_data.update(kwargs)
        self._queue.append(cast_data)


class RatingTest(tests.TestCase):
    def setUp(self):
        super(RatingTest, self).setUp()
        self._tenant_id = 'f266f30b11f246b589fd266f85eeec39'
        self._module = tests.FakeRatingModule(self._tenant_id)
        self._fake_rpc = FakeRPCClient()

    def test_get_module_info(self):
        mod_infos = self._module.module_info
        expected_infos = {'name': 'fake',
                          'description': 'fake rating module',
                          'hot_config': False,
                          'enabled': False,
                          'priority': 1}
        self.assertEqual(expected_infos, mod_infos)

    def test_set_state_triggers_rpc(self):
        with mock.patch('cloudkitty.messaging.get_client') as rpcmock:
            rpcmock.return_value = self._fake_rpc
            self._module.set_state(True)
            self.assertTrue(self._fake_rpc._fanout)
            self.assertEqual('rating', self._fake_rpc._namespace)
            self.assertEqual(1, len(self._fake_rpc._queue))
            rpc_data = self._fake_rpc._queue[0]
            expected_data = {'ctx': {},
                             'data': 'enable_module',
                             'name': 'fake'}
            self.assertEqual(expected_data, rpc_data)
            self._module.set_state(False)
            self.assertEqual(2, len(self._fake_rpc._queue))
            rpc_data = self._fake_rpc._queue[1]
            expected_data['data'] = 'disable_module'
            self.assertEqual(expected_data, rpc_data)

    def test_enable_module(self):
        with mock.patch('cloudkitty.messaging.get_client') as rpcmock:
            rpcmock.return_value = self._fake_rpc
            self._module.set_state(True)
        db_api = ck_db_api.get_instance()
        module_db = db_api.get_module_info()
        self.assertTrue(module_db.get_state('fake'))

    def test_disable_module(self):
        with mock.patch('cloudkitty.messaging.get_client') as rpcmock:
            rpcmock.return_value = self._fake_rpc
            self._module.set_state(False)
        db_api = ck_db_api.get_instance()
        module_db = db_api.get_module_info()
        self.assertFalse(module_db.get_state('fake'))

    def test_enabled_property(self):
        db_api = ck_db_api.get_instance()
        module_db = db_api.get_module_info()
        module_db.set_state('fake', True)
        self.assertTrue(self._module.enabled)
        module_db.set_state('fake', False)
        self.assertFalse(self._module.enabled)

    def test_get_default_priority(self):
        self.assertEqual(1, self._module.priority)

    def test_set_priority(self):
        self._module.set_priority(10)
        db_api = ck_db_api.get_instance()
        module_db = db_api.get_module_info()
        self.assertEqual(10, module_db.get_priority('fake'))

    def test_update_priority(self):
        old_prio = self._module.priority
        self._module.set_priority(10)
        new_prio = self._module.priority
        self.assertNotEqual(old_prio, new_prio)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import datetime
import decimal
import fractions
import itertools
import unittest

import mock
from oslo_utils import timeutils

from cloudkitty import utils as ck_utils


def iso2dt(iso_str):
    return timeutils.parse_isotime(iso_str)


class UtilsTimeCalculationsTest(unittest.TestCase):
    def setUp(self):
        self.date_ts = 1416219015
        self.date_iso = '2014-11-17T10:10:15Z'
        self.date_params = {'year': 2014,
                            'month': 11,
                            'day': 17,
                            'hour': 10,
                            'minute': 10,
                            'second': 15}
        self.date_tz_params = {'year': 2014,
                               'month': 10,
                               'day': 26,
                               'hour': 2,
                               'minute': 00,
                               'second': 00}

    def test_dt2ts(self):
        date = datetime.datetime(**self.date_params)
        trans_ts = ck_utils.dt2ts(date)
        self.assertEqual(self.date_ts, trans_ts)

    def test_iso2dt(self):
        date = datetime.datetime(**self.date_params)
        trans_dt = ck_utils.iso2dt(self.date_iso)
        self.assertEqual(date, trans_dt)

    def test_ts2iso(self):
        trans_iso = ck_utils.ts2iso(self.date_ts)
        self.assertEqual(self.date_iso, trans_iso)

    def test_dt2iso(self):
        date = datetime.datetime(**self.date_params)
        trans_iso = ck_utils.dt2iso(date)
        self.assertEqual(self.date_iso, trans_iso)

    @mock.patch.object(ck_utils, 'utcnow',
                       return_value=iso2dt('2014-01-31T00:00:00Z'))
    def test_month_start_without_dt(self, patch_utcnow_mock):
        date = datetime.datetime(2014, 1, 1)
        trans_dt = ck_utils.get_month_start()
        self.assertEqual(date, trans_dt)
        patch_utcnow_mock.assert_called_once_with()

    @mock.patch.object(ck_utils, 'utcnow',
                       return_value=iso2dt('2014-01-15T00:00:00Z'))
    def test_month_end_without_dt(self, patch_utcnow_mock):
        date = datetime.datetime(2014, 1, 31)
        trans_dt = ck_utils.get_month_end()
        self.assertEqual(date, trans_dt)
        patch_utcnow_mock.assert_called_once_with()

    @mock.patch.object(ck_utils, 'utcnow',
                       return_value=iso2dt('2014-01-31T00:00:00Z'))
    def test_get_last_month_without_dt(self, patch_utcnow_mock):
        date = datetime.datetime(2013, 12, 1)
        trans_dt = ck_utils.get_last_month()
        self.assertEqual(date, trans_dt)
        patch_utcnow_mock.assert_called_once_with()

    @mock.patch.object(ck_utils, 'utcnow',
                       return_value=iso2dt('2014-01-31T00:00:00Z'))
    def test_get_next_month_without_dt(self, patch_utcnow_mock):
        date = datetime.datetime(2014, 2, 1)
        trans_dt = ck_utils.get_next_month()
        self.assertEqual(date, trans_dt)
        patch_utcnow_mock.assert_called_once_with()

    def test_get_last_month_leap(self):
        base_date = datetime.datetime(2016, 3, 31)
        date = datetime.datetime(2016, 2, 1)
        trans_dt = ck_utils.get_last_month(base_date)
        self.assertEqual(date, trans_dt)

    def test_get_next_month_leap(self):
        base_date = datetime.datetime(2016, 1, 31)
        date = datetime.datetime(2016, 2, 1)
        trans_dt = ck_utils.get_next_month(base_date)
        self.assertEqual(date, trans_dt)

    def test_add_month_leap(self):
        base_date = datetime.datetime(2016, 1, 31)
        date = datetime.datetime(2016, 3, 3)
        trans_dt = ck_utils.add_month(base_date, False)
        self.assertEqual(date, trans_dt)

    def test_add_month_keep_leap(self):
        base_date = datetime.datetime(2016, 1, 31)
        date = datetime.datetime(2016, 2, 29)
        trans_dt = ck_utils.add_month(base_date)
        self.assertEqual(date, trans_dt)

    def test_sub_month_leap(self):
        base_date = datetime.datetime(2016, 3, 31)
        date = datetime.datetime(2016, 3, 3)
        trans_dt = ck_utils.sub_month(base_date, False)
        self.assertEqual(date, trans_dt)

    def test_sub_month_keep_leap(self):
        base_date = datetime.datetime(2016, 3, 31)
        date = datetime.datetime(2016, 2, 29)
        trans_dt = ck_utils.sub_month(base_date)
        self.assertEqual(date, trans_dt)

    def test_load_timestamp(self):
        calc_dt = ck_utils.iso2dt(self.date_iso)
        check_dt = ck_utils.ts2dt(self.date_ts)
        self.assertEqual(calc_dt, check_dt)


class ConvertUnitTest(unittest.TestCase):
    """Class testing the convert_unit and num2decimal function"""
    possible_args = [
        None,  # Use default arg
        '2/3',
        decimal.Decimal(1.23),
        '1.23',
        2,
        '2',
        2.3,
    ]

    def test_arg_types(self):
        """Test function with several arg combinations of different types"""
        for fac, off in itertools.product(self.possible_args, repeat=2):
            factor = fac if fac else 1
            offset = off if off else 0
            ck_utils.convert_unit(10, factor, offset)

    def test_str_str_str(self):
        result = ck_utils.convert_unit('1/2', '1/2', '1/2')
        self.assertEqual(result, decimal.Decimal(0.5 * 0.5 + 0.5))

    def test_str_float_float(self):
        result = ck_utils.convert_unit('1/2', 0.5, 0.5)
        self.assertEqual(result, decimal.Decimal(0.5 * 0.5 + 0.5))

    def test_convert_str_float(self):
        result = ck_utils.num2decimal('2.0')
        self.assertEqual(result, decimal.Decimal(2.0))

    def test_convert_str_int(self):
        result = ck_utils.num2decimal('2')
        self.assertEqual(result, decimal.Decimal(2))

    def test_convert_str_fraction(self):
        result = ck_utils.num2decimal('2/3')
        self.assertEqual(result, decimal.Decimal(2.0 / 3))

    def test_convert_fraction(self):
        result = ck_utils.num2decimal(fractions.Fraction(1, 2))
        self.assertEqual(result, decimal.Decimal(1.0 / 2))

    def test_convert_float(self):
        result = ck_utils.num2decimal(0.5)
        self.assertEqual(result, decimal.Decimal(0.5))

    def test_convert_int(self):
        result = ck_utils.num2decimal(2)
        self.assertEqual(result, decimal.Decimal(2))

    def test_convert_decimal(self):
        result = ck_utils.num2decimal(decimal.Decimal(2))
        self.assertEqual(result, decimal.Decimal(2))



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import os

import alembic
from alembic import config as alembic_config

ALEMBIC_INI_PATH = os.path.join(os.path.dirname(__file__), 'alembic.ini')


def load_alembic_config(repo_path, ini_path=ALEMBIC_INI_PATH):
    if not os.path.exists(repo_path):
        raise Exception('Repo path (%s) not found.' % repo_path)
    if not os.path.exists(ini_path):
        raise Exception('Ini path (%s) not found.' % ini_path)
    config = alembic_config.Config(ini_path)
    config.set_main_option('script_location', repo_path)
    return config


def upgrade(config, version):
    return alembic.command.upgrade(config, version or 'head')


def version(config):
    return alembic.command.current(config)


def revision(config, message='', autogenerate=False):
    """Creates template for migration.

    :param message: Text that will be used for migration title
    :type message: string
    :param autogenerate: If True - generates diff based on current database
                            state
    :type autogenerate: bool
    """
    return alembic.command.revision(config, message=message,
                                    autogenerate=autogenerate)


def stamp(config, revision):
    """Stamps database with provided revision.

    :param revision: Should match one from repository or head - to stamp
                        database with most recent revision
    :type revision: string
    """
    return alembic.command.stamp(config, revision=revision)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fixed UUID length problem.

Revision ID: d39836d70aee
Revises: 4c2f20df7491
Create Date: 2016-05-11 14:04:10.984006

"""

# revision identifiers, used by Alembic.
revision = 'd39836d70aee'
down_revision = '4c2f20df7491'

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('ghybrid_dataframes') as batch_op:
        batch_op.alter_column(
            'resource_ref',
            type_=sa.String(36),
            existing_type=sa.String(32))



# Copyright 2017 GohighSec.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from cloudkitty.common.policies import base

collector_policies = [
    policy.DocumentedRuleDefault(
        name='collector:list_mappings',
        check_str=base.ROLE_ADMIN,
        description='Return the list of every services mapped to a collector.',
        operations=[{'path': '/v1/collector/mappings',
                     'method': 'LIST'}]),
    policy.DocumentedRuleDefault(
        name='collector:get_mapping',
        check_str=base.ROLE_ADMIN,
        description='Return a service to collector mapping.',
        operations=[{'path': '/v1/collector/mappings/{service_id}',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='collector:manage_mapping',
        check_str=base.ROLE_ADMIN,
        description='Manage a service to collector mapping.',
        operations=[{'path': '/v1/collector/mappings',
                     'method': 'POST'},
                    {'path': '/v1/collector/mappings/{service_id}',
                     'method': 'DELETE'}]),
    policy.DocumentedRuleDefault(
        name='collector:get_state',
        check_str=base.ROLE_ADMIN,
        description='Query the enable state of a collector.',
        operations=[{'path': '/v1/collector/states/{collector_id}',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='collector:update_state',
        check_str=base.ROLE_ADMIN,
        description='Set the enable state of a collector.',
        operations=[{'path': '/v1/collector/states/{collector_id}',
                     'method': 'PUT'}])
]


def list_rules():
    return collector_policies
